# -*- coding: utf-8; -*-

import os

def countlines(start, lines=0, header=True, begin_start=None):
    if header:
        print('{:>10} |{:>10} | {:<20}'.format('ADDED', 'TOTAL', 'FILE'))
        print('{:->11}|{:->11}|{:->20}'.format('', '', ''))

    for thing in os.listdir(start):
        thing = os.path.join(start, thing)

        # With the following line only non-GUI code is counted.
        if os.path.isfile(thing) and not os.path.isfile(thing[:-3] + '.ui'):
        # As an alternative count all files.
        # if os.path.isfile(thing):
            if thing.endswith('.py'):
                with open(thing, 'r') as f:
                    newlines = f.readlines()
                    newlines = len(newlines)
                    lines += newlines

                    if begin_start is not None:
                        reldir_of_thing = '.' + thing.replace(begin_start, '')
                    else:
                        reldir_of_thing = '.' + thing.replace(start, '')

                    print('{:>10} |{:>10} | {:<20}'.format(
                            newlines, lines, reldir_of_thing))


    for thing in os.listdir(start):
        thing = os.path.join(start, thing)
        if os.path.isdir(thing):
            lines = countlines(thing, lines, header=False, begin_start=start)

    return lines

directory = r'D:\SW-Development\Python\PlanetarySystemStacker\planetary_system_stacker'
lines = countlines(directory)



import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from cv2 import VideoCapture, cvtColor, COLOR_RGB2GRAY
from numpy import uint8, uint16

from frames import detect_bayer, detect_rgb_bgr


def show_image(frame, comment):
    """
    Use Matplotlib to show a given frame.

    :param frame: Color or B/W image.
    :param comment: Text string (to be displayed as window title)
    :return: -
    """
    print(comment + ", shape: " + str(frame.shape) + ", type: " + str(frame.dtype))

    # If image is 16bit, reduce it to 8bit for display.
    if frame.dtype == uint16:
        frame_shown = (frame >> 8).astype(uint8)
    else:
        frame_shown = frame

    plt.title(comment)
    if len(frame_shown.shape) == 3:
        plt.imshow(frame_shown)
    else:
        plt.imshow(frame_shown, cmap='gray')
    plt.show()


filename = 'D:\SW-Development\Python\PlanetarySystemStacker\Examples\AVI_Chris-Garry\\20110929_005012_jupiter_gbrg_DIB.avi'
# filename = 'D:\SW-Development\Python\PlanetarySystemStacker\Examples\Saturn\sat_c11_ser_F0001-1731_short.avi'
# filename = 'D:/SW-Development/Python/PlanetarySystemStacker/Examples/Moon_2018-03-24/another_short_video.avi'

# Create the VideoCapture object.
cap = VideoCapture(filename)

# Read the first frame.
ret, first_frame_read = cap.read()
show_image(first_frame_read, "First frame of Video")

frames_bayer_max_noise_diff_green = 2.
frames_bayer_min_distance_from_blue = 99.5
frames_color_difference_threshold = 0
debayer_code = detect_bayer(first_frame_read, frames_bayer_max_noise_diff_green,
                            frames_bayer_min_distance_from_blue,
                            frames_color_difference_threshold)
print ("Debayer code found: " + debayer_code)

if debayer_code == 'Color':
    rgb_order = detect_rgb_bgr(first_frame_read)
    print ("Color order detected: " + rgb_order)


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'postproc_editor_gui.ui'
#
# Created by: PyQt5 UI code generator 5.13.0
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_postproc_editor(object):
    def setupUi(self, postproc_editor):
        postproc_editor.setObjectName("postproc_editor")
        postproc_editor.resize(865, 758)
        postproc_editor.setLocale(QtCore.QLocale(QtCore.QLocale.English, QtCore.QLocale.UnitedStates))
        postproc_editor.setFrameShape(QtWidgets.QFrame.Panel)
        postproc_editor.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.gridLayout = QtWidgets.QGridLayout(postproc_editor)
        self.gridLayout.setObjectName("gridLayout")
        self.VersionManagerWidget = QtWidgets.QLabel(postproc_editor)
        self.VersionManagerWidget.setText("")
        self.VersionManagerWidget.setObjectName("VersionManagerWidget")
        self.gridLayout.addWidget(self.VersionManagerWidget, 1, 2, 1, 2)
        self.tabWidget_postproc_control = QtWidgets.QTabWidget(postproc_editor)
        self.tabWidget_postproc_control.setObjectName("tabWidget_postproc_control")
        self.tab_sharpening = QtWidgets.QWidget()
        self.tab_sharpening.setObjectName("tab_sharpening")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.tab_sharpening)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.pushButton_add_layer = QtWidgets.QPushButton(self.tab_sharpening)
        self.pushButton_add_layer.setObjectName("pushButton_add_layer")
        self.verticalLayout_2.addWidget(self.pushButton_add_layer)
        self.scrollArea = QtWidgets.QScrollArea(self.tab_sharpening)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 400, 615))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.scrollAreaWidgetContents)
        self.verticalLayout.setObjectName("verticalLayout")
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.verticalLayout_2.addWidget(self.scrollArea)
        self.tabWidget_postproc_control.addTab(self.tab_sharpening, "")
        self.tab_rgb = QtWidgets.QWidget()
        self.tab_rgb.setObjectName("tab_rgb")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.tab_rgb)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.group_general = QtWidgets.QGroupBox(self.tab_rgb)
        self.group_general.setObjectName("group_general")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.group_general)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.checkBox_automatic = QtWidgets.QCheckBox(self.group_general)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.checkBox_automatic.sizePolicy().hasHeightForWidth())
        self.checkBox_automatic.setSizePolicy(sizePolicy)
        self.checkBox_automatic.setMinimumSize(QtCore.QSize(0, 0))
        self.checkBox_automatic.setObjectName("checkBox_automatic")
        self.verticalLayout_3.addWidget(self.checkBox_automatic)
        self.gridLayout_2 = QtWidgets.QGridLayout()
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.fgw_label_parameter = QtWidgets.QLabel(self.group_general)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.fgw_label_parameter.sizePolicy().hasHeightForWidth())
        self.fgw_label_parameter.setSizePolicy(sizePolicy)
        self.fgw_label_parameter.setMinimumSize(QtCore.QSize(0, 0))
        self.fgw_label_parameter.setObjectName("fgw_label_parameter")
        self.gridLayout_2.addWidget(self.fgw_label_parameter, 1, 0, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(36, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout_2.addItem(spacerItem, 1, 4, 1, 1)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout_2.addItem(spacerItem1, 1, 6, 1, 1)
        spacerItem2 = QtWidgets.QSpacerItem(40, 25, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout_2.addItem(spacerItem2, 1, 1, 1, 1)
        self.fgw_label_display = QtWidgets.QLabel(self.group_general)
        self.fgw_label_display.setObjectName("fgw_label_display")
        self.gridLayout_2.addWidget(self.fgw_label_display, 1, 5, 1, 1)
        self.label_resolution = QtWidgets.QLabel(self.group_general)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_resolution.sizePolicy().hasHeightForWidth())
        self.label_resolution.setSizePolicy(sizePolicy)
        self.label_resolution.setMinimumSize(QtCore.QSize(0, 0))
        self.label_resolution.setObjectName("label_resolution")
        self.gridLayout_2.addWidget(self.label_resolution, 0, 0, 1, 1)
        self.comboBox_resolution = QtWidgets.QComboBox(self.group_general)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.comboBox_resolution.sizePolicy().hasHeightForWidth())
        self.comboBox_resolution.setSizePolicy(sizePolicy)
        self.comboBox_resolution.setMinimumSize(QtCore.QSize(90, 0))
        self.comboBox_resolution.setObjectName("comboBox_resolution")
        self.gridLayout_2.addWidget(self.comboBox_resolution, 0, 2, 1, 1)
        spacerItem3 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout_2.addItem(spacerItem3, 0, 3, 1, 1)
        self.fgw_slider_value = QtWidgets.QSlider(self.group_general)
        self.fgw_slider_value.setMinimum(1)
        self.fgw_slider_value.setMaximum(6)
        self.fgw_slider_value.setSingleStep(1)
        self.fgw_slider_value.setPageStep(1)
        self.fgw_slider_value.setProperty("value", 4)
        self.fgw_slider_value.setSliderPosition(4)
        self.fgw_slider_value.setOrientation(QtCore.Qt.Horizontal)
        self.fgw_slider_value.setObjectName("fgw_slider_value")
        self.gridLayout_2.addWidget(self.fgw_slider_value, 1, 2, 1, 2)
        self.gridLayout_2.setColumnStretch(1, 1)
        self.gridLayout_2.setColumnStretch(2, 3)
        self.gridLayout_2.setColumnStretch(3, 8)
        self.gridLayout_2.setColumnStretch(4, 1)
        self.verticalLayout_3.addLayout(self.gridLayout_2)
        self.verticalLayout_3.setStretch(0, 1)
        self.verticalLayout_3.setStretch(1, 2)
        self.verticalLayout_4.addWidget(self.group_general)
        spacerItem4 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_4.addItem(spacerItem4)
        self.group_red = QtWidgets.QGroupBox(self.tab_rgb)
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.group_red.setFont(font)
        self.group_red.setObjectName("group_red")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.group_red)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.pushButton_red_up = QtWidgets.QPushButton(self.group_red)
        self.pushButton_red_up.setStyleSheet("")
        self.pushButton_red_up.setObjectName("pushButton_red_up")
        self.gridLayout_3.addWidget(self.pushButton_red_up, 1, 2, 1, 1)
        self.label_red_left = QtWidgets.QLabel(self.group_red)
        self.label_red_left.setStyleSheet("QLabel { color :red; }")
        self.label_red_left.setAlignment(QtCore.Qt.AlignCenter)
        self.label_red_left.setObjectName("label_red_left")
        self.gridLayout_3.addWidget(self.label_red_left, 2, 0, 1, 1)
        self.pushButton_red_down = QtWidgets.QPushButton(self.group_red)
        self.pushButton_red_down.setStyleSheet("")
        self.pushButton_red_down.setObjectName("pushButton_red_down")
        self.gridLayout_3.addWidget(self.pushButton_red_down, 3, 2, 1, 1)
        self.label_red_down = QtWidgets.QLabel(self.group_red)
        self.label_red_down.setStyleSheet("QLabel { color :red; }")
        self.label_red_down.setAlignment(QtCore.Qt.AlignCenter)
        self.label_red_down.setObjectName("label_red_down")
        self.gridLayout_3.addWidget(self.label_red_down, 4, 2, 1, 1)
        self.pushButton_red_reset = QtWidgets.QPushButton(self.group_red)
        self.pushButton_red_reset.setObjectName("pushButton_red_reset")
        self.gridLayout_3.addWidget(self.pushButton_red_reset, 2, 2, 1, 1)
        self.label_red_right = QtWidgets.QLabel(self.group_red)
        self.label_red_right.setStyleSheet("QLabel { color :red; }")
        self.label_red_right.setAlignment(QtCore.Qt.AlignCenter)
        self.label_red_right.setObjectName("label_red_right")
        self.gridLayout_3.addWidget(self.label_red_right, 2, 4, 1, 1)
        self.label_red_up = QtWidgets.QLabel(self.group_red)
        self.label_red_up.setLayoutDirection(QtCore.Qt.LeftToRight)
        self.label_red_up.setStyleSheet("QLabel { color :red; }")
        self.label_red_up.setAlignment(QtCore.Qt.AlignCenter)
        self.label_red_up.setObjectName("label_red_up")
        self.gridLayout_3.addWidget(self.label_red_up, 0, 2, 1, 1)
        self.pushButton_red_left = QtWidgets.QPushButton(self.group_red)
        self.pushButton_red_left.setStyleSheet("")
        self.pushButton_red_left.setObjectName("pushButton_red_left")
        self.gridLayout_3.addWidget(self.pushButton_red_left, 2, 1, 1, 1)
        self.pushButton_red_right = QtWidgets.QPushButton(self.group_red)
        self.pushButton_red_right.setStyleSheet("")
        self.pushButton_red_right.setObjectName("pushButton_red_right")
        self.gridLayout_3.addWidget(self.pushButton_red_right, 2, 3, 1, 1)
        self.verticalLayout_4.addWidget(self.group_red)
        self.group_blue = QtWidgets.QGroupBox(self.tab_rgb)
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.group_blue.setFont(font)
        self.group_blue.setObjectName("group_blue")
        self.gridLayout_4 = QtWidgets.QGridLayout(self.group_blue)
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.label_blue_up = QtWidgets.QLabel(self.group_blue)
        self.label_blue_up.setStyleSheet("QLabel { color :blue; }")
        self.label_blue_up.setAlignment(QtCore.Qt.AlignCenter)
        self.label_blue_up.setObjectName("label_blue_up")
        self.gridLayout_4.addWidget(self.label_blue_up, 0, 2, 1, 1)
        self.pushButton_blue_up = QtWidgets.QPushButton(self.group_blue)
        self.pushButton_blue_up.setStyleSheet("")
        self.pushButton_blue_up.setObjectName("pushButton_blue_up")
        self.gridLayout_4.addWidget(self.pushButton_blue_up, 1, 2, 1, 1)
        self.label_blue_left = QtWidgets.QLabel(self.group_blue)
        self.label_blue_left.setStyleSheet("QLabel { color :blue; }")
        self.label_blue_left.setAlignment(QtCore.Qt.AlignCenter)
        self.label_blue_left.setObjectName("label_blue_left")
        self.gridLayout_4.addWidget(self.label_blue_left, 2, 0, 1, 1)
        self.pushButton_blue_left = QtWidgets.QPushButton(self.group_blue)
        self.pushButton_blue_left.setStyleSheet("")
        self.pushButton_blue_left.setObjectName("pushButton_blue_left")
        self.gridLayout_4.addWidget(self.pushButton_blue_left, 2, 1, 1, 1)
        self.pushButton_blue_reset = QtWidgets.QPushButton(self.group_blue)
        self.pushButton_blue_reset.setObjectName("pushButton_blue_reset")
        self.gridLayout_4.addWidget(self.pushButton_blue_reset, 2, 2, 1, 1)
        self.pushButton_blue_right = QtWidgets.QPushButton(self.group_blue)
        self.pushButton_blue_right.setStyleSheet("")
        self.pushButton_blue_right.setObjectName("pushButton_blue_right")
        self.gridLayout_4.addWidget(self.pushButton_blue_right, 2, 3, 1, 1)
        self.label_blue_right = QtWidgets.QLabel(self.group_blue)
        self.label_blue_right.setStyleSheet("QLabel { color :blue; }")
        self.label_blue_right.setAlignment(QtCore.Qt.AlignCenter)
        self.label_blue_right.setObjectName("label_blue_right")
        self.gridLayout_4.addWidget(self.label_blue_right, 2, 4, 1, 1)
        self.pushButton_blue_down = QtWidgets.QPushButton(self.group_blue)
        self.pushButton_blue_down.setStyleSheet("")
        self.pushButton_blue_down.setObjectName("pushButton_blue_down")
        self.gridLayout_4.addWidget(self.pushButton_blue_down, 3, 2, 1, 1)
        self.label_blue_down = QtWidgets.QLabel(self.group_blue)
        self.label_blue_down.setStyleSheet("QLabel { color :blue; }")
        self.label_blue_down.setAlignment(QtCore.Qt.AlignCenter)
        self.label_blue_down.setObjectName("label_blue_down")
        self.gridLayout_4.addWidget(self.label_blue_down, 4, 2, 1, 1)
        self.verticalLayout_4.addWidget(self.group_blue)
        self.verticalLayout_4.setStretch(0, 8)
        self.verticalLayout_4.setStretch(1, 1)
        self.verticalLayout_4.setStretch(2, 8)
        self.verticalLayout_4.setStretch(3, 8)
        self.tabWidget_postproc_control.addTab(self.tab_rgb, "")
        self.gridLayout.addWidget(self.tabWidget_postproc_control, 0, 2, 1, 2)
        self.FrameViewer = QtWidgets.QLabel(postproc_editor)
        self.FrameViewer.setText("")
        self.FrameViewer.setAlignment(QtCore.Qt.AlignCenter)
        self.FrameViewer.setObjectName("FrameViewer")
        self.gridLayout.addWidget(self.FrameViewer, 0, 1, 1, 1)
        self.label_zoom = QtWidgets.QLabel(postproc_editor)
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.label_zoom.setFont(font)
        self.label_zoom.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label_zoom.setObjectName("label_zoom")
        self.gridLayout.addWidget(self.label_zoom, 2, 1, 1, 1)
        self.label_message = QtWidgets.QLabel(postproc_editor)
        self.label_message.setObjectName("label_message")
        self.gridLayout.addWidget(self.label_message, 2, 0, 1, 1)
        self.buttonBox = QtWidgets.QDialogButtonBox(postproc_editor)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.gridLayout.addWidget(self.buttonBox, 2, 2, 1, 1)
        self.gridLayout.setColumnStretch(0, 4)
        self.gridLayout.setColumnStretch(1, 1)

        self.retranslateUi(postproc_editor)
        self.tabWidget_postproc_control.setCurrentIndex(0)
        self.checkBox_automatic.toggled['bool'].connect(self.group_red.setDisabled)
        self.checkBox_automatic.toggled['bool'].connect(self.group_blue.setDisabled)
        QtCore.QMetaObject.connectSlotsByName(postproc_editor)

    def retranslateUi(self, postproc_editor):
        _translate = QtCore.QCoreApplication.translate
        postproc_editor.setWindowTitle(_translate("postproc_editor", "Frame"))
        self.pushButton_add_layer.setToolTip(_translate("postproc_editor", "Add a sharpening layer (up to four)"))
        self.pushButton_add_layer.setText(_translate("postproc_editor", "Add correction layer"))
        self.tabWidget_postproc_control.setTabText(self.tabWidget_postproc_control.indexOf(self.tab_sharpening), _translate("postproc_editor", "Wavelets"))
        self.group_general.setTitle(_translate("postproc_editor", "RGB Alignment parameters for the version selected"))
        self.checkBox_automatic.setToolTip(_translate("postproc_editor", "Perform RGB Alignment in fully automatic mode.\n"
"This option is available in batch mode as well."))
        self.checkBox_automatic.setText(_translate("postproc_editor", "Automatic"))
        self.fgw_label_parameter.setToolTip(_translate("postproc_editor", "In the presence of noise in the image, shift detection between frames can be misled by spurious local minima.\n"
"Gaussian blur can help finding the global optimum. Try higher values for noisy images."))
        self.fgw_label_parameter.setText(_translate("postproc_editor", "Noise level"))
        self.fgw_label_display.setText(_translate("postproc_editor", "TextLabel"))
        self.label_resolution.setText(_translate("postproc_editor", "Resolution"))
        self.comboBox_resolution.setToolTip(_translate("postproc_editor", "Choose (sub-)pixel resolution for RGB alignment. Please note\n"
"that RAM usage and execution time increase quadratically."))
        self.fgw_slider_value.setToolTip(_translate("postproc_editor", "Smoothen the image before computing the channel shifts.\n"
"In the presence of noise this results in better accuracy."))
        self.group_red.setToolTip(_translate("postproc_editor", "The current red channel shift is shown.\n"
"Use push buttons for manual adjustments."))
        self.group_red.setTitle(_translate("postproc_editor", "Red channel status / manual adjustments"))
        self.pushButton_red_up.setToolTip(_translate("postproc_editor", "Shift the red channel upwards."))
        self.pushButton_red_up.setText(_translate("postproc_editor", "Up"))
        self.label_red_left.setText(_translate("postproc_editor", "0.25"))
        self.pushButton_red_down.setToolTip(_translate("postproc_editor", "Shift the red channel down."))
        self.pushButton_red_down.setText(_translate("postproc_editor", "Down"))
        self.label_red_down.setText(_translate("postproc_editor", "0.25"))
        self.pushButton_red_reset.setToolTip(_translate("postproc_editor", "Reset the red channel shift to zero."))
        self.pushButton_red_reset.setText(_translate("postproc_editor", "Reset"))
        self.label_red_right.setText(_translate("postproc_editor", "0.25"))
        self.label_red_up.setText(_translate("postproc_editor", "0.25"))
        self.pushButton_red_left.setToolTip(_translate("postproc_editor", "Shift the red channel left."))
        self.pushButton_red_left.setText(_translate("postproc_editor", "Left"))
        self.pushButton_red_right.setToolTip(_translate("postproc_editor", "Shift the red channel right."))
        self.pushButton_red_right.setText(_translate("postproc_editor", "Right"))
        self.group_blue.setToolTip(_translate("postproc_editor", "The current blue channel shift is shown.\n"
"Use push buttons for manual adjustments."))
        self.group_blue.setTitle(_translate("postproc_editor", "Blue channel status / manual adjustments"))
        self.label_blue_up.setText(_translate("postproc_editor", "0.25"))
        self.pushButton_blue_up.setToolTip(_translate("postproc_editor", "Shift the blue channel up."))
        self.pushButton_blue_up.setText(_translate("postproc_editor", "Up"))
        self.label_blue_left.setText(_translate("postproc_editor", "0.25"))
        self.pushButton_blue_left.setToolTip(_translate("postproc_editor", "Shift the blue channel left."))
        self.pushButton_blue_left.setText(_translate("postproc_editor", "Left"))
        self.pushButton_blue_reset.setToolTip(_translate("postproc_editor", "Reset the blue channel shift to zero."))
        self.pushButton_blue_reset.setText(_translate("postproc_editor", "Reset"))
        self.pushButton_blue_right.setToolTip(_translate("postproc_editor", "Shift the blue channel right."))
        self.pushButton_blue_right.setText(_translate("postproc_editor", "Right"))
        self.label_blue_right.setText(_translate("postproc_editor", "0.25"))
        self.pushButton_blue_down.setToolTip(_translate("postproc_editor", "Shift the blue channel down."))
        self.pushButton_blue_down.setText(_translate("postproc_editor", "Down"))
        self.label_blue_down.setText(_translate("postproc_editor", "0.25"))
        self.tabWidget_postproc_control.setTabText(self.tabWidget_postproc_control.indexOf(self.tab_rgb), _translate("postproc_editor", "RGB Alignment"))
        self.label_zoom.setText(_translate("postproc_editor", "100%"))
        self.label_message.setText(_translate("postproc_editor", "message"))
        self.buttonBox.setToolTip(_translate("postproc_editor", "Press \"OK\" to save the selected version and exit, or \"Cancel\" to discard changes"))



# -*- coding: utf-8; -*-
"""
Copyright (c) 2018 Rolf Hempel, rolf6419@gmx.de

This file is part of the PlanetarySystemStacker tool (PSS).
https://github.com/Rolf-Hempel/PlanetarySystemStacker

PSS is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PSS.  If not, see <http://www.gnu.org/licenses/>.

"""

from glob import glob
from time import time

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from math import ceil
from numpy import arange, amax, stack, amin, float32, uint8, zeros, sqrt, empty, int32, uint16
from scipy import ndimage
try:
    from skimage.registration import phase_cross_correlation
except ImportError:
    from skimage.feature import register_translation as phase_cross_correlation
from cv2 import meanStdDev, GaussianBlur

from align_frames import AlignFrames
from configuration import Configuration
from exceptions import NotSupportedError, Error
from frames import Frames
from miscellaneous import Miscellaneous
from rank_frames import RankFrames


class AlignmentPoints(object):
    """
        Create a rectangular grid of potential places for alignment points. For each location
        create a small "alignment box" around it. For each alignment box test if there is enough
        structure and brightness in the picture to use it as an alignment point. For all
        alignment points in all frames, compute the local shifts relative to the mean frame.

    """

    def __init__(self, configuration, frames, rank_frames, align_frames, progress_signal=None):
        """
        Initialize the AlignmentPoints object and compute the mean frame.

        :param configuration: Configuration object with parameters
        :param frames: Frames object with all video frames
        :param rank_frames: RankFrames object with global quality ranks (between 0. and 1.,
                            1. being optimal) for all frames
        :param align_frames: AlignFrames object with global shift information for all frames
        :param progress_signal: Either None (no progress signalling), or a signal with the signature
                                (str, int) with the current activity (str) and the progress in
                                percent (int).
        """

        self.configuration = configuration
        self.frames = frames
        self.rank_frames = rank_frames
        self.align_frames = align_frames
        self.progress_signal = progress_signal

        # Apply a low-pass filter on the mean frame as a preparation for shift detection.
        self.mean_frame = GaussianBlur(align_frames.mean_frame.astype(uint16),
                                        (self.configuration.frames_gauss_width,
                                        self.configuration.frames_gauss_width), 0).astype(int32)
        self.num_pixels_y = self.mean_frame.shape[0]
        self.num_pixels_x = self.mean_frame.shape[1]
        self.alignment_points_dropped_dim = None
        self.alignment_points_dropped_structure = None

        # Initialize counters for APs which are dropped because
        # they do not satisfy the brightness or structure condition.
        self.alignment_points = []
        self.alignment_points_dropped_dim = 0
        self.alignment_points_dropped_structure = 0

        # Initialize the number of frames to be stacked at each AP.
        self.stack_size = None

        self.dev_table = empty((2 * self.configuration.alignment_points_search_width + 1,
                               2 * self.configuration.alignment_points_search_width + 1), dtype=float32)

    @staticmethod
    def ap_locations(num_pixels, min_boundary_distance, step_size, even):
        """
        Compute optimal alignment patch coordinates in one coordinate direction. Place boundary
        neighbors as close as possible to the boundary.

        :param num_pixels: Number of pixels in the given coordinate direction
        :param min_boundary_distance: Minimum distance of an AP from the boundary
        :param step_size: Distance of alignment boxes
        :param even: If True, compute locations for even row indices. Otherwise, for odd ones.
        :return: List of alignment point coordinates in the given direction
        """
        # Initialize list of locations.
        locations = []

        # If the frame size is too small for the AP size chosen, no AP locations can be computed.
        # In this case an empty list is returned.
        try:
            # The number of interior alignment boxes in general is not an integer. Round to the next
            # higher number.
            num_interior_odd = int(ceil((num_pixels - 2 * min_boundary_distance) / step_size))
            # Because alignment points are arranged in a staggered grid, in even rows there is one point
            # more.
            num_interior_even = num_interior_odd + 1

            # The precise distance between alignment points will differ slightly from the specified
            # step_size. Compute the exact distance. Integer locations will be rounded later.
            distance_corrected = (num_pixels - 2 * min_boundary_distance) / num_interior_odd

            # Compute the AP locations, separately for even and odd rows.
            if even:
                locations = [int(min_boundary_distance + i * distance_corrected) for i in range(num_interior_even)]
            else:
                locations = [int(min_boundary_distance + 0.5 * distance_corrected +
                                 i * distance_corrected) for i in range(num_interior_odd)]
        except:
            pass
        return locations

    def create_ap_grid(self):
        """
        Create a 2D staggered grid of alignment points. For each AP compute its center coordinates,
        and the coordinate limits of its alignment box and alignment patch. Only alignment points
        which satisfy the conditions on brightness, contrast and structure are eventually added to
        the list.

        :return: List of alignment points.
        """

        # The alignment patch is the area which is stacked after a rigid displacement.
        half_patch_width = self.configuration.alignment_points_half_patch_width
        # The alignment box is the object on which the displacement computation is performed.
        half_box_width = self.configuration.alignment_points_half_box_width
        # Number of pixels in one coordinate direction between alignment points
        step_size = self.configuration.alignment_points_step_size
        # Maximum displacement searched for in the alignment process.
        search_width = self.configuration.alignment_points_search_width
        # Minimum structure value for an alignment point (between 0. and 1.)
        structure_threshold = self.configuration.alignment_points_structure_threshold
        # The brightest pixel must be brighter than this value (0 < value <256). Please note that
        # brightness and contrast values are converted to 16bit resolution.
        brightness_threshold = self.configuration.alignment_points_brightness_threshold * 256
        # The difference between the brightest and darkest pixel values must be larger than this
        # value (0 < value < 256)
        contrast_threshold = self.configuration.alignment_points_contrast_threshold * 256

        # Reset the alignment point list, and initialize counters for APs which are dropped because
        # they do not satisfy the brightness or structure condition.
        self.alignment_points = []
        self.alignment_points_dropped_dim = 0
        self.alignment_points_dropped_structure = 0

        # Compute the minimum distance of an AP from the boundary.
        min_boundary_distance = max(half_box_width + search_width, half_patch_width)

        # Compute y and x coordinate locations of alignment points. Note that the grid is staggered.
        ap_locations_y = self.ap_locations(self.num_pixels_y, min_boundary_distance,
                                           step_size, True)
        ap_locations_x_even = self.ap_locations(self.num_pixels_x, min_boundary_distance,
                                                step_size, True)
        ap_locations_x_odd = self.ap_locations(self.num_pixels_x, min_boundary_distance,
                                               step_size, False)

        # If no AP coordinates fit into the frame, no APs are created.
        if not ap_locations_y or not ap_locations_x_even or not ap_locations_x_odd:
            return

        # Compute the minimum distance of an AP center from the frame boundary.
        min_boundary_distance = max(
            self.configuration.alignment_points_half_box_width + \
            self.configuration.alignment_points_search_width,
            self.configuration.alignment_points_half_patch_width)

        # Create alignment point rows, start with an even one.
        even = True
        for index_y, y in enumerate(ap_locations_y):
            # For the first row extend the patch to the upper frame border, and for the last row
            # to the lower frame border.
            extend_y_low  = (index_y == 0)
            extend_y_high = (index_y == len(ap_locations_y)-1)

            # Create x coordinate, depending on the y row being even or odd (staggered grid).
            if even:
                ap_locations_x = ap_locations_x_even
            else:
                ap_locations_x = ap_locations_x_odd

            # For each location create an alignment point.
            for index_x, x in enumerate(ap_locations_x):
                # For the first point in a row, extend the patch to the left frame border, and for
                # the last point in a row to the right frame border.
                extend_x_low  = (index_x == 0)
                extend_x_high = (index_x == len(ap_locations_x) - 1)

                alignment_point = self.new_alignment_point(y, x, extend_x_low, extend_x_high,
                                                           extend_y_low, extend_y_high)

                # Compute structure and brightness information for the alignment box.
                max_brightness = amax(alignment_point['reference_box'])
                min_brightness = amin(alignment_point['reference_box'])
                # If the alignment box satisfies the brightness conditions, add the AP to the list.
                if max_brightness > brightness_threshold and max_brightness - \
                        min_brightness > contrast_threshold:

                    # Check if the fraction of dark pixels exceeds a threshold.
                    box = alignment_point['reference_box']
                    fraction = (box < brightness_threshold).sum() / (box.shape[0] * box.shape[1])
                    if fraction > self.configuration.alignment_points_dim_fraction_threshold:

                        # Compute the center of mass of the brightness distribution within the box,
                        # and shift the box center to this location.
                        com = ndimage.measurements.center_of_mass(box)
                        y_adapted = y + int(com[0]) - half_box_width
                        x_adapted = x + int(com[1]) - half_box_width

                        y_adapted = max(y_adapted, min_boundary_distance)
                        y_adapted = min(y_adapted, self.num_pixels_y - min_boundary_distance)
                        x_adapted = max(x_adapted, min_boundary_distance)
                        x_adapted = min(x_adapted, self.num_pixels_x - min_boundary_distance)

                        # Replace the alignment point with a new one, using the updated
                        # coordinates.
                        alignment_point = self.new_alignment_point(y_adapted, x_adapted,
                                                                   extend_x_low, extend_x_high,
                                                                   extend_y_low, extend_y_high)

                    alignment_point['structure'] = Miscellaneous.quality_measure(
                        alignment_point['reference_box'])
                    self.alignment_points.append(alignment_point)
                else:
                    # If a point does not satisfy the conditions, increase the counter.
                    self.alignment_points_dropped_dim += 1

            # Switch between even and odd rows.
            even = not even

        # Normalize the structure information for all alignment point boxes by dividing by the
        # maximum value.
        if self.alignment_points:
            structure_max = max(
                alignment_point['structure'] for alignment_point in self.alignment_points)
            alignment_points_dropped_structure_indices = []
            for alignment_point_index, alignment_point in enumerate(self.alignment_points):
                alignment_point['structure'] /= structure_max
                # Remove alignment points with too little structure and increment the counter.
                if alignment_point['structure'] < structure_threshold:
                    alignment_points_dropped_structure_indices.append(alignment_point_index)
                    self.alignment_points_dropped_structure += 1

            # Remove alignment points which do not satisfy the structure condition, if there is any.
            if alignment_points_dropped_structure_indices:
                alignment_points_new = []
                dropped_index = 0
                for alignment_point_index, alignment_point in enumerate(self.alignment_points):
                    if alignment_point_index != alignment_points_dropped_structure_indices[
                        dropped_index]:
                        alignment_points_new.append(alignment_point)
                    elif dropped_index < len(alignment_points_dropped_structure_indices) - 1:
                        dropped_index += 1
                self.alignment_points = alignment_points_new

    def new_alignment_point(self, y, x, extend_x_low, extend_x_high, extend_y_low, extend_y_high):
        """
        Create a new alignment point. This method is called in creating the initial alignment point
        grid. Later it can be invoked by the user to add single alignment points.

        :param y: y coordinate of alignment point center
        :param x: x coordinate of alignment point center
        :param extend_x_low: True, if patch is to be extended to the left frame boundary.
                             False otherwise.
        :param extend_x_high: True, if patch is to be extended to the right frame boundary.
                              False otherwise.
        :param extend_y_low: True, if patch is to be extended to the upper frame boundary.
                             False otherwise.
        :param extend_y_high: True, if patch is to be extended to the lower frame boundary.
                              False otherwise.
        :return: If successful, the alignment point object is returned; otherwise None.
        """

        # If the patch does not fit into the frame, or the AP box is too close to the border,
        # the AP cannot be created at this place.
        min_boundary_distance = max(self.configuration.alignment_points_half_box_width + \
                                self.configuration.alignment_points_search_width,
                                self.configuration.alignment_points_half_patch_width)
        if y<min_boundary_distance or y>self.num_pixels_y-min_boundary_distance or \
            x<min_boundary_distance or x>self.num_pixels_x-min_boundary_distance:
            return None

        alignment_point = {}
        alignment_point['y'] = y
        alignment_point['x'] = x
        alignment_point['half_box_width'] = self.configuration.alignment_points_half_box_width
        alignment_point['box_y_low'] = y - self.configuration.alignment_points_half_box_width
        alignment_point['box_y_high'] = y + self.configuration.alignment_points_half_box_width
        alignment_point['box_x_low'] = x - self.configuration.alignment_points_half_box_width
        alignment_point['box_x_high'] = x + self.configuration.alignment_points_half_box_width

        alignment_point['patch_y_low'] = y - self.configuration.alignment_points_half_patch_width
        alignment_point['patch_y_high'] = y + self.configuration.alignment_points_half_patch_width
        if extend_y_low:
            alignment_point['patch_y_low'] = 0
        elif extend_y_high:
            alignment_point['patch_y_high'] = self.num_pixels_y

        alignment_point['patch_x_low'] = x - self.configuration.alignment_points_half_patch_width
        alignment_point['patch_x_high'] = x + self.configuration.alignment_points_half_patch_width
        if extend_x_low:
            alignment_point['patch_x_low'] = 0
        elif extend_x_high:
            alignment_point['patch_x_high'] = self.num_pixels_x

        # Initialize the reference to the corresponding widget in the AP viewer scene.
        alignment_point['graphics_item'] = None

        # Allocate buffers and fill alignment point box with mean frame.
        AlignmentPoints.set_reference_box(alignment_point, self.mean_frame)

        return alignment_point

    def add_alignment_point(self, ap):
        """
        Add an alignment point to the list.

        :param ap: AP to be added
        :return: -
        """

        self.alignment_points.append(ap)

    def remove_alignment_points(self, ap_list):
        """
        Remove a list of alignment points from the current list.

        :param ap_list: List of point objects to be removed
        :return: -
        """

        # Create list with unique identifiers of all items on the list.
        ap_list_ids = [id(ap_list_item) for ap_list_item in ap_list]

        # Replace the original AP list with the reduced one.
        # If the identifier of an alignment point does not match any list item, keep it.
        self.alignment_points = [ap for ap in self.alignment_points if id(ap) not in ap_list_ids]

    def replace_alignment_point(self, ap_old, ap_new):
        """
        Replace an alignment point with a new one.

        :param ap_old: Existing alignment point to be replaced
        :param ap_new: New alignment point to replace the old one
        :return: True if successful, False otherwise
        """

        ap_old_id = id(ap_old)
        for index, ap in enumerate(self.alignment_points):
            if id(ap) == ap_old_id:
                self.alignment_points[index] = ap_new
                return True
        return False

    def move_alignment_point(self, ap, y_new, x_new):
        """
        Move an existing AP to a different position.

        :param ap: Esisting alignment point to be moved
        :param y_new: New y coordinate of AP center
        :param x_new: New x coordinate of AP center
        :return: The updated AP; or None, if unsuccessful
        """

        # Try to do the changes. If unsuccessful, return None.
        return self.change_alignment_point(ap, y_new, x_new, ap['half_box_width'])

    def resize_alignment_point(self, ap, factor):
        """
        Change the size of an existing alignment point.

        :param ap: Esisting alignment point to be resized
        :param factor: Factor by which the size is to be changed
        :return: The resized alignment point; or None, if unsuccessful
        """

        half_box_width_new = ap['half_box_width'] * factor

        # Try to do the changes. If unsuccessful, return None.
        return self.change_alignment_point(ap, ap["y"], ap["x"], half_box_width_new)

    def change_alignment_point(self, ap, y, x, half_box_width):
        """
        Try to fit a resized or moved AP into the frame.

        :param ap: AP to be resized or moved
        :param y: New y coordinate of center
        :param x: New x coordinate of center
        :param half_box_width: New half width of AP box
        :return: The resized / moved AP; or None, if unsuccessful
        """

        # Test if the AP box is too small.
        if half_box_width < self.configuration.alignment_points_min_half_box_width:
            return None

        # Compute new values for patch and box sizes.
        half_patch_width_new_int = round(half_box_width *
                                             self.configuration.alignment_points_half_patch_width /
                                             self.configuration.alignment_points_half_box_width)
        half_box_width_new_int = round(half_box_width)

        # Compute resized patch bounds. If resizing hits the image boundary on at least one side,
        # the operation is aborted.
        patch_y_low = y - half_patch_width_new_int
        if patch_y_low < 0:
            return None
        patch_y_high = y + half_patch_width_new_int
        if patch_y_high > self.num_pixels_y:
            return None
        patch_x_low = x - half_patch_width_new_int
        if patch_x_low < 0:
            return None
        patch_x_high = x + half_patch_width_new_int
        if patch_x_high > self.num_pixels_x:
            return None

        # Compute resized box bounds. If resizing moves the box closer to the frame border than
        # the search width on at least one side, the operation is aborted.
        box_y_low = y - half_box_width_new_int
        if box_y_low < self.configuration.alignment_points_search_width:
            return None
        box_y_high = y + half_box_width_new_int
        if box_y_high > self.num_pixels_y - self.configuration.alignment_points_search_width:
            return None
        box_x_low = x - half_box_width_new_int
        if box_x_low < self.configuration.alignment_points_search_width:
            return None
        box_x_high = x + half_box_width_new_int
        if box_x_high > self.num_pixels_x - self.configuration.alignment_points_search_width:
            return None

        # Perform the changes.
        ap['y'] = y
        ap['x'] = x
        ap['patch_y_low'] = patch_y_low
        ap['patch_y_high'] = patch_y_high
        ap['patch_x_low'] = patch_x_low
        ap['patch_x_high'] = patch_x_high
        ap['box_y_low'] = box_y_low
        ap['box_y_high'] = box_y_high
        ap['box_x_low'] = box_x_low
        ap['box_x_high'] = box_x_high
        ap['half_box_width'] = half_box_width

        # Invalidate buffers. To save computing time, they are not re-computed at every wheel
        # event.
        ap['graphics_item'] = None
        ap['reference_box'] = None
        ap['stacking_buffer'] = None

        return ap

    @staticmethod
    def set_reference_box(alignment_point, mean_frame):
        """
        For APs which have been changed in the AP editor, buffers have been invalidated. They have
        to be re-computed after AP editing is done.

        :param alignment_point: Alignment_point object
        :param mean_frame: Average frame
        :return: -
        """

        # Cut out the reference box from the mean frame, used in alignment.
        box = mean_frame[alignment_point['box_y_low']:alignment_point['box_y_high'],
              alignment_point['box_x_low']:alignment_point['box_x_high']]
        alignment_point['reference_box'] = box

    def set_reference_boxes_correlation(self):
        """
        This method is used if multilevel-correlation AP matching is selected. In this case two
        reference boxes with different resolution are constructed. The box used in the first phase
        contains only every second pixel of the original frame in each coordinate direction. The box
        used in the second phase has the same resolution as the original frame.

        This method is invoked when all APs are set, just before stacking.

        :return: -
        """

        for ap_index, alignment_point in enumerate(self.alignment_points):

            # Cut the reference box for this alignment point from the mean frame.
            window_second_phase = self.align_frames.mean_frame[
                                  alignment_point['box_y_low']:
                                  alignment_point['box_y_high'],
                                  alignment_point['box_x_low']:
                                  alignment_point['box_x_high']].astype(float32)

            # The reference box with the full resolution is used in the second phase.
            alignment_point['reference_box_second_phase'] = window_second_phase

            # In the first phase a box with half the resolution is constructed.
            alignment_point['reference_box_first_phase'] =  window_second_phase[::2, ::2]

    @staticmethod
    def initialize_ap_stacking_buffer(alignment_point, drizzle_factor, color):
        """
        In the stacking initialization, for each AP a stacking buffer has to be allocated. At the
        same time, drizzled patch index bounds are computed.

        :param alignment_point: Alignment_point object
        :param drizzle_factor: Drizzle factor (integer: 1, 2 or 3)
        :param color: True, if stacking is to be done for color frames. False for monochrome case.
        :return: -
        """

        # Compute drizzled patch coordinates and index bounds.
        alignment_point['y_drizzled'] = alignment_point['y'] * drizzle_factor
        alignment_point['x_drizzled'] = alignment_point['x'] * drizzle_factor
        alignment_point['patch_y_low_drizzled'] = alignment_point['patch_y_low'] * drizzle_factor
        alignment_point['patch_y_high_drizzled'] = alignment_point['patch_y_high'] * drizzle_factor
        alignment_point['patch_x_low_drizzled'] = alignment_point['patch_x_low'] * drizzle_factor
        alignment_point['patch_x_high_drizzled'] = alignment_point['patch_x_high'] * drizzle_factor

        # Allocate space for the stacking buffer.
        if color:
            alignment_point['stacking_buffer'] = zeros(
                [alignment_point['patch_y_high_drizzled'] - alignment_point['patch_y_low_drizzled'],
                 alignment_point['patch_x_high_drizzled'] - alignment_point['patch_x_low_drizzled'],
                 3],
                dtype=float32)
        else:
            alignment_point['stacking_buffer'] = zeros(
                [alignment_point['patch_y_high_drizzled'] - alignment_point['patch_y_low_drizzled'],
                 alignment_point['patch_x_high_drizzled'] - alignment_point[
                     'patch_x_low_drizzled']], dtype=float32)

    def find_alignment_points(self, y_low, y_high, x_low, x_high):
        """
        Find all alignment points the centers of which are within given (y, x) bounds.

        :param y_low: Lower y pixel coordinate bound
        :param y_high: Upper y pixel coordinate bound
        :param x_low: Lower x pixel coordinate bound
        :param x_high: Upper x pixel coordinate bound
        :return: List of all alignment points with centers within the given coordinate bounds.
                 If no AP satisfies the condition, return an empty list.
        """

        return [ap for ap in self.alignment_points if y_low <= ap['y'] <= y_high and
                x_low <= ap['x'] <= x_high]

    @staticmethod
    def find_neighbor(ap_y, ap_x, alignment_points):
        """
        For a given (y, x) position find the closest "real" alignment point.

        :param ap_y: y cocrdinate of location of interest
        :param ap_x: x cocrdinate of location of interest
        :param alignment_points: List of alignment points to be searched

        :return: Alignment point object of closest AP, and distance in pixels. If the list of
                 alignment points is empty, (None, None) is returned.
        """

        # If no APs have been created yet, return None for both results.
        if not alignment_points:
            return None, None

        # Start with an impossibly large distance, and find the closest AP on the list.
        min_distance_squared = 1.e30
        ap_neighbor = None
        for ap in alignment_points:
            distance_squared = (ap['y'] - ap_y) ** 2 + (ap['x'] - ap_x) ** 2
            if distance_squared < min_distance_squared:
                ap_neighbor = ap
                min_distance_squared = distance_squared
        return ap_neighbor, sqrt(min_distance_squared)

    def compute_frame_qualities(self):
        """
        For each alignment point compute a ranking of best frames. Store the list in the
        alignment point dictionary with the key 'best_frame_indices'.

        Consider the special case that sampled-down Laplacians have been stored for frame ranking.
        In this case they can be re-used for ranking the boxes around alignment points (but only
        if "Laplace" has been selected for alignment point ranking).

        :return: -
        """

        # If the user has entered a value for the number of frames, use it.
        if self.configuration.alignment_points_frame_number > 0:
            self.stack_size = self.configuration.alignment_points_frame_number
        # Otherwise compute the stack size from the given percentage. Take at least one frame.
        else:
            self.stack_size = max(int(ceil(
                self.frames.number * self.configuration.alignment_points_frame_percent / 100.)), 1)

        # Select the ranking method.
        if self.configuration.alignment_points_rank_method == "xy gradient":
            method = Miscellaneous.local_contrast
        elif self.configuration.alignment_points_rank_method == "Laplace":
            method = Miscellaneous.local_contrast_laplace
        elif self.configuration.alignment_points_rank_method == "Sobel":
            method = Miscellaneous.local_contrast_sobel
        else:
            raise NotSupportedError(
                "Ranking method " + self.configuration.alignment_points_rank_method +
                " not supported")

        # Compute the frequency of progress signals in the computational loop.
        if self.progress_signal is not None:
            self.signal_loop_length = max(self.frames.number, 1)
            self.signal_step_size = max(round(self.frames.number / 10), 1)

        # Initialize a list which for each AP contains the qualities of all frames at this point.
        for alignment_point in self.alignment_points:
            alignment_point['frame_qualities'] = []

        if self.configuration.rank_frames_method != "Laplace" or \
                self.configuration.alignment_points_rank_method != "Laplace":
            # There are no stored Laplacians, or they cannot be used for the specified method.
            # Cycle through all frames and alignment points:
            for frame_index in range(self.frames.number):
                frame = self.frames.frames_mono_blurred(frame_index)

                # After every "signal_step_size"th frame, send a progress signal to the main GUI.
                if self.progress_signal is not None and frame_index % self.signal_step_size == 1:
                    self.progress_signal.emit("Rank frames at APs",
                                    int(round(10 * frame_index / self.signal_loop_length) * 10))

                for ap_index, alignment_point in enumerate(self.alignment_points):
                    # Compute patch bounds within the current frame.
                    y_low = max(0,
                                alignment_point['patch_y_low'] + self.align_frames.dy[frame_index])
                    y_high = min(self.frames.shape[0],
                                 alignment_point['patch_y_high'] + self.align_frames.dy[
                                     frame_index])
                    x_low = max(0,
                                alignment_point['patch_x_low'] + self.align_frames.dx[frame_index])
                    x_high = min(self.frames.shape[1],
                                 alignment_point['patch_x_high'] + self.align_frames.dx[
                                     frame_index])
                    # Compute the frame quality and append it to the list for this alignment point.
                    if self.configuration.frames_normalization:
                        alignment_point['frame_qualities'].append(
                            method(frame[y_low:y_high, x_low:x_high],
                                   self.configuration.alignment_points_rank_pixel_stride) /
                            self.frames.average_brightness(frame_index))
                    else:
                        alignment_point['frame_qualities'].append(
                            method(frame[y_low:y_high, x_low:x_high],
                                   self.configuration.alignment_points_rank_pixel_stride))
        else:
            # Sampled-down Laplacians of all blurred frames have been computed in
            # "frames.frames_mono_blurred_laplacian". Cut out boxes around alignment points from
            # those objects, rather than computing new Laplacians. Cycle through all frames and
            # alignment points. Use the blurred monochrome image for ranking.
            for frame_index in range(self.frames.number):
                frame = self.frames.frames_mono_blurred_laplacian(frame_index)

                # After every "signal_step_size"th frame, send a progress signal to the main GUI.
                if self.progress_signal is not None and frame_index % self.signal_step_size == 1:
                    self.progress_signal.emit("Rank frames at APs",
                                        int(round(10 * frame_index / self.signal_loop_length) * 10))

                for ap_index, alignment_point in enumerate(self.alignment_points):
                    # Compute patch bounds within the current frame.
                    y_low = int(max(0, alignment_point['patch_y_low'] + self.align_frames.dy[
                        frame_index]) / self.configuration.align_frames_sampling_stride)
                    y_high = int(min(self.frames.shape[0],
                                     alignment_point['patch_y_high'] + self.align_frames.dy[
                                         frame_index]) / self.configuration.align_frames_sampling_stride)
                    x_low = int(max(0, alignment_point['patch_x_low'] + self.align_frames.dx[
                        frame_index]) / self.configuration.align_frames_sampling_stride)
                    x_high = int(min(self.frames.shape[1],
                                     alignment_point['patch_x_high'] + self.align_frames.dx[
                                         frame_index]) / self.configuration.align_frames_sampling_stride)
                    # Compute the frame quality and append it to the list for this alignment point.
                    if self.configuration.frames_normalization:
                        alignment_point['frame_qualities'].append(
                            meanStdDev(frame[y_low:y_high, x_low:x_high])[1][0][0] /
                            self.frames.average_brightness(frame_index))
                    else:
                        alignment_point['frame_qualities'].append(
                            meanStdDev(frame[y_low:y_high, x_low:x_high])[1][0][0])

        if self.progress_signal is not None:
            self.progress_signal.emit("Rank frames at APs", 100)

        # Initialize the alignment point lists for all frames.
        self.frames.reset_alignment_point_lists()
        # For each alignment point sort the computed quality ranks in descending order.
        for alignment_point_index, alignment_point in enumerate(self.alignment_points):
            # Truncate the list to the number of frames to be stacked for each alignmeent point.
            alignment_point['best_frame_indices'] = sorted(range(len(alignment_point['frame_qualities'])),
                                                    key=alignment_point['frame_qualities'].__getitem__, reverse=True)[:self.stack_size]
            # Add this alignment point to the AP lists of those frames where the AP is to be used.
            for frame_index in alignment_point['best_frame_indices']:
                self.frames.used_alignment_points[frame_index].append(alignment_point_index)

    def compute_shift_alignment_point(self, frame_mono_blurred, frame_index, alignment_point_index,
                                      de_warp=True, weight_matrix_first_phase=None,
                                      subpixel_solve=False):
        """
        Compute the [y, x] pixel shift vector at a given alignment point relative to the mean frame.
        Four different methods can be used to compute the shift values:
        - a multilevel cross correlation algorithm (miscellaneous.multilevel_correlation). This
          method was implemented after PSS had shown bad performance for planetary videos captured
          with bad seeing. Pixel shifts can be computed with subpixel accuracy, as specified with
          the parameter "subpixel_solve". This is needed for drizzling.
        - a subpixel algorithm from "skimage.feature"
        - a phase correlation algorithm (miscellaneous.translation)
        - a local search algorithm (spiralling outwards), see method "search_local_match",
          optionally with subpixel accuracy.
        - a local search algorithm, based on steepest descent, see method
          "search_local_match_gradient". This method is faster than the previous one, but it has no
          subpixel option.

        Be careful with the sign of the local shift values. For the first two methods, a positive
        value means that the current frame has to be shifted in the positive coordinate direction
        in order to make objects in the frame match with their counterparts in the reference frame.
        In other words: If the shift is positive, an object in the current frame is at lower pixel
        coordinates as compared to the reference. This is very counter-intuitive, but to make the
        three methods consistent, the same approach was followed in the implementation of the third
        method "search_local_match", contained in this module. There, a pixel box around an
        alignment point in the current frame is moved until the content of the box matches with the
        corresponding box in the reference frame. If at this point the box is shifted towards a
        higher coordinate value, this value is returned with a negative sign as the local shift.

        :param frame_mono_blurred: Gaussian-blurred version of the frame with index "frame_index"
        :param frame_index: Index of the selected frame in the list of frames
        :param alignment_point_index: Index of the selected alignment point
        :param de_warp: If True, include local warp shift computation. If False, only apply
                        global frame shift.
        :param weight_matrix_first_phase: This parameter is only used by the alignment method
                                          "MultiLevelCorrelation". If not None it defines a
                                          weighting array by which the cross correlation results are
                                          multiplied before the maximum value is determined. The
                                          size of this 2D array in each coordinate direction is that
                                          of the reference_box_first_phase plus two times the first
                                          phase search width.
        :param subpixel_solve: Used only for alignment method "MultiLevelCorrelation". If True,
                               in the second phase the optimum is computed with sub-pixel accuracy
                               (i.e. the returned shifts are not integer). If False, shifts are
                               computed as integer values.

        :return: [shift_y, shift_x], success with: [shift_y, shift_x]: Local shift vector
                                         success: True, if computation successful; False, otherwise.
        """

        # Initialize the "success" return value.
        success = True

        alignment_point = self.alignment_points[alignment_point_index]
        y_low = alignment_point['box_y_low']
        y_high = alignment_point['box_y_high']
        x_low = alignment_point['box_x_low']
        x_high = alignment_point['box_x_high']

        # The offsets dy and dx are caused by two effects: First, the mean frame is smaller
        # than the original frames. It only contains their intersection. And second, because the
        # given frame is globally shifted as compared to the mean frame.
        dy = self.align_frames.dy[frame_index]
        dx = self.align_frames.dx[frame_index]

        if de_warp:
            # Use a two-level algorithm based on (weighted) cross correlation. The user-supplied
            # noise level parameter is used as the blurring factor for the first correlation phase.
            if self.configuration.alignment_points_method == 'MultiLevelCorrelation':
                shift_y_local_first_phase, shift_x_local_first_phase, success_first_phase, \
                shift_y_local_second_phase, shift_x_local_second_phase, success_second_phase = \
                    Miscellaneous.multilevel_correlation(
                        alignment_point['reference_box_first_phase'],
                        frame_mono_blurred, self.configuration.frames_gauss_width,
                        alignment_point['reference_box_second_phase'],
                        y_low + dy, y_high + dy, x_low + dx, x_high + dx,
                        self.configuration.alignment_points_search_width,
                        weight_matrix_first_phase=weight_matrix_first_phase,
                        subpixel_solve=subpixel_solve)

                # Compute the combined warp shifts from both phases.
                shift_pixel = [shift_y_local_first_phase + shift_y_local_second_phase,
                               shift_x_local_first_phase + shift_x_local_second_phase]
                success = success_second_phase

            # Use subpixel registration from skimage.feature, with accuracy 1/10 pixels.
            elif self.configuration.alignment_points_method == 'Subpixel':
                # Cut out the alignment box from the given frame. Take into account the offsets
                # explained above.
                box_in_frame = frame_mono_blurred[y_low + dy:y_high + dy,
                               x_low + dx:x_high + dx]
                # The following use of "phase_cross_correlation" replaces the original call to
                # "register_translation" which seems not to be in the package anymore. If this
                # replacement really works has not been tested.
                shift_pixel, error, diffphase = phase_cross_correlation(
                    alignment_point['reference_box'], box_in_frame, upsample_factor=10)

            # Use a simple phase shift computation (contained in module "miscellaneous").
            elif self.configuration.alignment_points_method == 'CrossCorrelation':
                # Cut out the alignment box from the given frame. Take into account the offsets
                # explained above.
                box_in_frame = frame_mono_blurred[y_low + dy:y_high + dy,
                                                  x_low + dx:x_high + dx]
                shift_pixel = Miscellaneous.translation(alignment_point['reference_box'],
                                                        box_in_frame, box_in_frame.shape)

            # Use a local search (see method "search_local_match" below.
            elif self.configuration.alignment_points_method == 'RadialSearch':
                shift_pixel, dev_r = Miscellaneous.search_local_match(alignment_point['reference_box'],
                    frame_mono_blurred, y_low + dy, y_high + dy, x_low + dx, x_high + dx,
                    self.configuration.alignment_points_search_width,
                    self.configuration.alignment_points_sampling_stride,
                    sub_pixel=self.configuration.alignment_points_local_search_subpixel)
                # If a zero shift was returned after a search with radius>2, thie means that
                # the search was not successfu.
                success = len(dev_r)<=2 or shift_pixel!=[0, 0]

            # Use the steepest descent search method.
            elif self.configuration.alignment_points_method == 'SteepestDescent':
                shift_pixel, dev_r = Miscellaneous.search_local_match_gradient(
                    alignment_point['reference_box'],
                    frame_mono_blurred, y_low + dy, y_high + dy, x_low + dx, x_high + dx,
                    self.configuration.alignment_points_search_width,
                    self.configuration.alignment_points_sampling_stride, self.dev_table)
                success = len(dev_r)<=2 or shift_pixel!=[0, 0]
            else:
                raise NotSupportedError("The point shift computation method " +
                                        self.configuration.alignment_points_method +
                                        " is not implemented")

            # Return the computed shift vector.
            return shift_pixel, success
        else:
            # If no de-warping is computed, just return the zero vector.
            return [0, 0], success

    def show_alignment_points(self, image):
        """
        Create an RGB version of a monochrome image and insert red crosses at all alignment
        point locations. Draw green alignment point boxes and white alignment point patches.

        :return: 8-bit RGB image with annotations.
        """

        if len(image.shape) == 3:
            color_image = image.astype(uint8)
        else:
            # Expand the monochrome reference frame to RGB
            color_image = stack((image.astype(uint8),) * 3, -1)

        # For all alignment boxes insert a color-coded cross.
        cross_half_len = 5

        for alignment_point in (self.alignment_points):
            y_center = alignment_point['y']
            x_center = alignment_point['x']
            Miscellaneous.insert_cross(color_image, y_center,
                                       x_center, cross_half_len, 'red')
            box_y_low = max(alignment_point['box_y_low'], 0)
            box_y_high = min(alignment_point['box_y_high'], image.shape[0]) - 1
            box_x_low = max(alignment_point['box_x_low'], 0)
            box_x_high = min(alignment_point['box_x_high'], image.shape[1]) - 1
            for y in arange(box_y_low, box_y_high):
                color_image[y, box_x_low] = [255, 255, 255]
                color_image[y, box_x_high] = [255, 255, 255]
            for x in arange(box_x_low, box_x_high):
                color_image[box_y_low, x] = [255, 255, 255]
                color_image[box_y_high, x] = [255, 255, 255]

            patch_y_low = max(alignment_point['patch_y_low'], 0)
            patch_y_high = min(alignment_point['patch_y_high'], image.shape[0]) - 1
            patch_x_low = max(alignment_point['patch_x_low'], 0)
            patch_x_high = min(alignment_point['patch_x_high'], image.shape[1]) - 1
            for y in arange(patch_y_low, patch_y_high):
                color_image[y, patch_x_low] = [0, int(
                    (255 + color_image[y, patch_x_low][1]) / 2.), 0]
                color_image[y, patch_x_high] = [0, int(
                    (255 + color_image[y, patch_x_high][1]) / 2.), 0]
            for x in arange(patch_x_low, patch_x_high):
                color_image[patch_y_low, x] = [0, int(
                    (255 + color_image[patch_y_low, x][1]) / 2.), 0]
                color_image[patch_y_high, x] = [0, int(
                    (255 + color_image[patch_y_high, x][1]) / 2.), 0]

        return color_image


if __name__ == "__main__":
    # Images can either be extracted from a video file or a batch of single photographs. Select
    # the example for the test run.
    type = 'video'
    if type == 'image':
        names = glob('Images/2012*.tif')
        # names = glob.glob('Images/Moon_Tile-031*ap85_8b.tif')
        # names = glob.glob('Images/Example-3*.jpg')
    else:
        names = 'Videos/short_video.avi'
    print(names)

    # Get configuration parameters.
    configuration = Configuration()
    configuration.initialize_configuration()
    try:
        frames = Frames(configuration, names, type=type)
        print("Number of images read: " + str(frames.number))
        print("Image shape: " + str(frames.shape))
    except Error as e:
        print("Error: " + e.message)
        exit()

    # Rank the frames by their overall local contrast.
    rank_frames = RankFrames(frames, configuration)
    start = time()
    rank_frames.frame_score()
    end = time()
    print('Elapsed time in ranking images: {}'.format(end - start))
    print("Index of maximum: " + str(rank_frames.frame_ranks_max_index))
    print("Frame scores: " + str(rank_frames.frame_ranks))
    print("Frame scores (sorted): " + str(
        [rank_frames.frame_ranks[i] for i in rank_frames.quality_sorted_indices]))
    print("Sorted index list: " + str(rank_frames.quality_sorted_indices))

    # Initialize the frame alignment object.
    align_frames = AlignFrames(frames, rank_frames, configuration)

    if configuration.align_frames_mode == "Surface":
        start = time()
        # Select the local rectangular patch in the image where the L gradient is highest in both x
        # and y direction. The scale factor specifies how much smaller the patch is compared to the
        # whole image frame.
        (y_low_opt, y_high_opt, x_low_opt, x_high_opt) = align_frames.compute_alignment_rect(
            configuration.align_frames_rectangle_scale_factor)
        end = time()
        print('Elapsed time in computing optimal alignment rectangle: {}'.format(end - start))
        print("optimal alignment rectangle, x_low: " + str(x_low_opt) + ", x_high: " + str(
            x_high_opt) + ", y_low: " + str(y_low_opt) + ", y_high: " + str(y_high_opt))
        reference_frame_with_alignment_points = frames.frames_mono(
            align_frames.frame_ranks_max_index).copy()
        reference_frame_with_alignment_points[y_low_opt,
        x_low_opt:x_high_opt] = reference_frame_with_alignment_points[y_high_opt - 1,
                                x_low_opt:x_high_opt] = 255
        reference_frame_with_alignment_points[y_low_opt:y_high_opt,
        x_low_opt] = reference_frame_with_alignment_points[y_low_opt:y_high_opt,
                     x_high_opt - 1] = 255
        # plt.imshow(reference_frame_with_alignment_points, cmap='Greys_r')
        # plt.show()

    # Align all frames globally relative to the frame with the highest score.
    start = time()
    align_frames.align_frames()
    end = time()
    print('Elapsed time in aligning all frames: {}'.format(end - start))
    print("Frame shifts: " + str(align_frames.frame_shifts))
    print("Intersection: " + str(align_frames.intersection_shape))

    start = time()
    # Compute the reference frame by averaging the best frames.
    average = align_frames.average_frame()

    # Initialize the AlignmentPoints object. This includes the computation of the average frame
    # against which the alignment point shifts are measured.

    alignment_points = AlignmentPoints(configuration, frames, rank_frames, align_frames)
    end = time()
    print('Elapsed time in computing reference frame: {}'.format(end - start))
    print("Reference frame computed from the best " + str(
        align_frames.average_frame_number) + " frames.")
    # plt.imshow(align_frames.mean_frame, cmap='Greys_r')
    # plt.show()

    # Create alignment points, and show alignment point boxes and patches.
    alignment_points.create_ap_grid()
    print("Number of alignment points created: " + str(len(alignment_points.alignment_points)) +
          ", number of dropped aps (dim): " + str(
        alignment_points.alignment_points_dropped_dim) +
          ", number of dropped aps (structure): " + str(
        alignment_points.alignment_points_dropped_structure))
    color_image = alignment_points.show_alignment_points(average)

    plt.imshow(color_image)
    plt.show()

    # For each alignment point rank frames by their quality.
    start = time()
    alignment_points.compute_frame_qualities()
    end = time()
    print('Elapsed time in ranking frames for every alignment point: {}'.format(end - start))

    y_low = 490
    y_high = 570
    x_low = 880
    x_high = 960
    found_ap_list = alignment_points.find_alignment_points(y_low, y_high, x_low, x_high)
    print("Removing alignment points between bounds " + str(y_low) + " <= y <= " + str(y_high) +
          ", " + str(x_low) + " <= x <= " + str(x_high) + ":")
    for ap in found_ap_list:
        print("y: " + str(ap['y']) + ", x: " + str(ap['x']))
    alignment_points.remove_alignment_points(found_ap_list)

    y_new = 530
    x_new = 920
    half_box_width_new = 40
    half_patch_width_new = 50
    num_pixels_y = average.shape[0]
    num_pixels_x = average.shape[1]
    alignment_points.alignment_points.append(
        alignment_points.new_alignment_point(y_new, x_new, False, False, False, False))
    print("Added alignment point at y: " + str(y_new) + ", x: " + str(x_new) + ", box size: "
          + str(2 * half_box_width_new) + ", patch size: " + str(2 * half_patch_width_new))

    # Show updated alignment point boxes and patches.
    print("Number of alignment points created: " + str(len(alignment_points.alignment_points)) +
          ", number of dropped aps (dim): " + str(alignment_points.alignment_points_dropped_dim) +
          ", number of dropped aps (structure): " + str(
        alignment_points.alignment_points_dropped_structure))
    color_image = alignment_points.show_alignment_points(average)

    plt.imshow(color_image)
    plt.show()



from argparse import ArgumentParser, ArgumentTypeError
from time import sleep
from glob import glob

from PyQt5 import QtCore

from configuration import Configuration
from exceptions import InternalError
from job_editor import Job
from miscellaneous import Miscellaneous
from workflow import Workflow

# Definition of data types, including value bounds, used in command line argument parsing.
def ram_size_type(x):
    try:
        x = int(x)
    except:
        raise ArgumentTypeError("Maximum RAM size must be an integer > 0")
    if x < 1:
        raise ArgumentTypeError("Maximum RAM size must be an integer > 0")
    return x


def noise_type(x):
    x = int(x)
    if not 0 <= x <= 11:
        raise ArgumentTypeError("Noise level must be between 0 and 11")
    return x


def stab_size_type(x):
    x = int(x)
    if not 5 <= x <= 80:
        raise ArgumentTypeError("Stabilization patch size must be between 5% and 80%")
    return x


def stab_sw_type(x):
    x = int(x)
    if not 5 <= x <= 150:
        raise ArgumentTypeError(
            "Stabilization search width must be between 5 and 150 pixels")
    return x


def rf_percent_type(x):
    x = int(x)
    if not 3 <= x <= 30:
        raise ArgumentTypeError(
            "Percentage of best frames for reference frame computation must be between 3% and 30%")
    return x


def align_box_width_type(x):
    x = int(x)
    if not 20 <= x <= 140:
        raise ArgumentTypeError(
            "Alignment point box width must be between 20 and 140 pixels")
    return x


def align_search_width_type(x):
    x = int(x)
    if not 6 <= x <= 30:
        raise ArgumentTypeError(
            "Alignment point search width must be between 6 and 30 pixels")
    return x


def align_min_struct_type(x):
    x = float(x)
    if not 0.01 <= x <= 0.30:
        raise ArgumentTypeError(
            "Alignment point minimum structure must be between 0.01 and 0.30")
    return x


def align_min_bright_type(x):
    x = int(x)
    if not 2 <= x <= 50:
        raise ArgumentTypeError(
            "Alignment point minimum brightness must be between 2 and 50")
    return x


def stack_percent_type(x):
    x = int(x)
    if not 1 <= x <= 100:
        raise ArgumentTypeError(
            "Percentage of best frames to be stacked must be between 1 and 100")
    return x


def stack_number_type(x):
    x = int(x)
    if not 1 <= x:
        raise ArgumentTypeError(
            "Number of best frames to be stacked must be greater or equal 1")
    return x


def normalize_bco_type(x):
    x = int(x)
    if not 0 <= x <= 40:
        raise ArgumentTypeError(
            "Normalization black cut-off must be between 0 and 40")
    return x


class PssConsole(QtCore.QObject):
    """
    This class replaces the class PlanetarySystemStacker if the program is started from the
    command line. In this case no GUI activity is created, and there is no interactive mode.
    """

    # Define signals which trigger activities on the workflow thread.
    signal_load_master_dark = QtCore.pyqtSignal(str)
    signal_load_master_flat = QtCore.pyqtSignal(str)
    signal_frames = QtCore.pyqtSignal(object)
    signal_rank_frames = QtCore.pyqtSignal(bool)
    signal_align_frames = QtCore.pyqtSignal(int, int, int, int)
    signal_set_roi = QtCore.pyqtSignal(int, int, int, int)
    signal_set_alignment_points = QtCore.pyqtSignal()
    signal_compute_frame_qualities = QtCore.pyqtSignal()
    signal_stack_frames = QtCore.pyqtSignal()
    signal_save_stacked_image = QtCore.pyqtSignal()

    def __init__(self, parent=None):
        super(PssConsole, self).__init__(parent)

        # Create the configuration object and modify it as specified in command line arguments.
        self.setup_configuration()

        # Start the workflow.
        self.work_next_task("Read frames")

    def setup_configuration(self):
        """
        Parse the command line arguments, initialize the configuration object and update
        configuration parameters with values passed via command line arguments.

        :return: -
        """
        parser = ArgumentParser()
        parser.add_argument("job_input", nargs='+', help="input video files or still image folders")

        parser.add_argument("-p", "--protocol", action="store_true",
                            help="Store protocol with results")
        parser.add_argument("--protocol_detail", type=int, choices=[0, 1, 2], default=1,
                            help="Protocol detail level")
        parser.add_argument("-b", "--buffering_level", choices=["auto", "0", "1", "2", "3", "4"],
                            default="auto", help="Buffering level")
        parser.add_argument("-r", "--ram_size", type=ram_size_type, default=-1,
                            help="Maximum RAM for this job (GBytes)")
        parser.add_argument("--out_format", choices=["png", "tiff", "fits"], default="png",
                            help="Image format for output")
        parser.add_argument("--name_add_f", action="store_true",
                            help="Add number of stacked frames to output file name")
        parser.add_argument("--name_add_p", action="store_true",
                            help="Add percentage of stacked frames to output file name")
        parser.add_argument("--name_add_apb", action="store_true",
                            help="Add alignment point box size (pixels) to output file name")
        parser.add_argument("--name_add_apn", action="store_true",
                            help="Add number of alignment points to output file name")

        parser.add_argument("--debayering",
                            choices=["Auto detect color", "Grayscale", "RGB", "RGB", "BGR",
                                     "Force Bayer RGGB", "Force Bayer GRBG",
                                     "Force Bayer GBRG", "Force Bayer BGGR"],
                            default="Auto detect color", help="Debayering option")
        parser.add_argument("--debayer_method",
                            choices=["Bilinear", "Variable Number of Gradients",
                                     "Edge Aware"],
                            default="Bilinear", help="Debayering method to be used")
        parser.add_argument("--noise", type=noise_type, default=7,
                            help="Noise level (add Gaussian blur)")
        parser.add_argument("-m", "--stab_mode", choices=["Surface", "Planet"], default="Surface",
                            help="Frame stabilization mode")
        parser.add_argument("--stab_size", type=stab_size_type, default=33,
                            help="Stabilization patch size (%% of frame)")
        parser.add_argument("--stab_sw", type=stab_sw_type, default=34,
                            help="Stabilization search width (pixels)")
        parser.add_argument("--rf_percent", type=rf_percent_type, default=5,
                            help="Percentage of best frames for reference frame computation")
        parser.add_argument("--fast_changing_object", action="store_true",
                            help="The object is changing fast during video time span (e.g. Jupiter")
        parser.add_argument("-d", "--dark", help="Image file for dark frame correction")
        parser.add_argument("-f", "--flat", help="Image file for flat frame correction")

        parser.add_argument("-a", "--align_box_width", type=align_box_width_type, default=48,
                            help="Alignment point box width (pixels)")
        parser.add_argument("-w", "--align_search_width", type=align_search_width_type, default=14,
                            help="Alignment point search width (pixels)")
        parser.add_argument("--align_min_struct", type=align_min_struct_type, default=0.04,
                            help="Alignment point minimum structure")
        parser.add_argument("--align_min_bright", type=align_min_bright_type, default=10,
                            help="Alignment point minimum brightness")

        parser.add_argument("-s", "--stack_percent", type=stack_percent_type, default=10,
                            help="Percentage of best frames to be stacked")
        parser.add_argument("--stack_number", type=stack_number_type,
                            help="Number of best frames to be stacked")
        parser.add_argument("-n", "--normalize_bright", action="store_true",
                            help="Normalize frame brightness")
        parser.add_argument("--normalize_bco", type=normalize_bco_type, default=15,
                            help="Normalization black cut-off")
        parser.add_argument("--drizzle", choices=["Off", "1.5x", "2x", "3x"], default="Off",
                            help="Drizzle factor (Off, 1.5x, 2x, 3x)")

        arguments = parser.parse_args()
        # self.print_arguments(arguments)

        # Create and initialize the configuration object. The configuration stored in the .ini file
        # in the user's home directory is ignored in this case. Modifications to standard values
        # come as command line arguments.
        self.configuration = Configuration()
        self.configuration.initialize_configuration(read_from_file=False)

        # In the standard configuration postprocessing is included in the workflow. This does not
        # make sense in command line mode.
        self.configuration.global_parameters_include_postprocessing = False

        # Modify the standard configuration as specified in the command line arguments.
        self.configuration.global_parameters_store_protocol_with_result = arguments.protocol
        self.configuration.global_parameters_protocol_level = arguments.protocol_detail
        if arguments.ram_size == -1:
            self.configuration.global_parameters_maximum_memory_active = False
            if arguments.buffering_level == "auto":
                self.configuration.global_parameters_buffering_level = -1
            else:
                self.configuration.global_parameters_buffering_level = int(arguments.buffering_level)
        else:
            self.configuration.global_parameters_buffering_level = -1
            self.configuration.global_parameters_maximum_memory_active = True
            self.configuration.global_parameters_maximum_memory_amount = arguments.ram_size

        self.configuration.global_parameters_image_format = arguments.out_format
        self.configuration.global_parameters_parameters_in_filename = arguments.name_add_f or \
            arguments.name_add_p or arguments.name_add_apb or arguments.name_add_apn
        self.configuration.global_parameters_stack_number_frames = arguments.name_add_f
        self.configuration.global_parameters_stack_percent_frames = arguments.name_add_p
        self.configuration.global_parameters_ap_box_size = arguments.name_add_apb
        self.configuration.global_parameters_ap_number = arguments.name_add_apn

        self.configuration.frames_debayering_default = arguments.debayering
        self.configuration.frames_debayering_method = arguments.debayer_method
        self.configuration.frames_gauss_width = arguments.noise

        self.configuration.align_frames_mode = arguments.stab_mode
        self.configuration.align_frames_rectangle_scale_factor = 100. / arguments.stab_size
        self.configuration.align_frames_search_width = arguments.stab_sw
        self.configuration.align_frames_average_frame_percent = arguments.rf_percent
        self.configuration.align_frames_fast_changing_object = arguments.fast_changing_object

        self.configuration.alignment_points_half_box_width = int(
            round(arguments.align_box_width / 2))
        self.configuration.alignment_points_search_width = arguments.align_search_width
        self.configuration.alignment_points_structure_threshold = arguments.align_min_struct
        self.configuration.alignment_points_brightness_threshold = arguments.align_min_bright

        self.configuration.alignment_points_frame_percent = arguments.stack_percent
        self.configuration.alignment_points_frame_number = -1

        # If the number of frames to be stacked is given, it has precedence over the percentage.
        if arguments.stack_number is not None:
            self.configuration.alignment_points_frame_number = arguments.stack_number
            self.configuration.alignment_points_frame_percent = -1

        self.configuration.frames_normalization = arguments.normalize_bright
        self.configuration.frames_normalization_threshold = arguments.normalize_bco
        self.configuration.stack_frames_drizzle_factor_string = arguments.drizzle

        # Re-compute derived parameters after the configuration was changed.
        self.configuration.set_derived_parameters()

        # Create the workflow thread and start it.
        self.thread = QtCore.QThread()
        self.workflow = Workflow(self)
        self.workflow.setParent(None)
        self.workflow.moveToThread(self.thread)
        self.workflow.calibration.report_calibration_error_signal.connect(
            self.report_calibration_error)
        self.workflow.work_next_task_signal.connect(self.work_next_task)
        self.workflow.report_error_signal.connect(self.report_error)
        self.workflow.abort_job_signal.connect(self.next_job_after_error)
        self.thread.start()

        # Connect signals to start activities on the workflow thread (e.g. in method
        # "work_next_task").
        self.signal_load_master_dark.connect(self.workflow.calibration.load_master_dark)
        self.signal_load_master_flat.connect(self.workflow.calibration.load_master_flat)
        self.signal_frames.connect(self.workflow.execute_frames)
        self.signal_rank_frames.connect(self.workflow.execute_rank_frames)
        self.signal_align_frames.connect(self.workflow.execute_align_frames)
        self.signal_set_roi.connect(self.workflow.execute_set_roi)
        self.signal_set_alignment_points.connect(self.workflow.execute_set_alignment_points)
        self.signal_compute_frame_qualities.connect(
            self.workflow.execute_compute_frame_qualities)
        self.signal_stack_frames.connect(self.workflow.execute_stack_frames)
        self.signal_save_stacked_image.connect(self.workflow.execute_save_stacked_image)

        # Set "automatic" to True. There is no interactive mode in this case.
        self.automatic = True

        # Create the job objects using the names passed as positional arguments.
        self.jobs = []
        for name in [f for name in arguments.job_input for f in glob(name)]:
            try:
                job = Job(name)

                # Test if the path specifies a stacking job.
                if job.type == 'video' or job.type == 'image':
                    # Override the "Auto detect color" value of the "Job" object with the
                    # command line value.
                    job.bayer_option_selected = self.configuration.frames_debayering_default
                    self.jobs.append(job)
                else:
                    if self.configuration.global_parameters_protocol_level > 0:
                        Miscellaneous.protocol(
                            "Error: '" + name + "' does not contain valid input for a stacking job,"
                                               " continune with next job.\n",
                            self.workflow.attached_log_file)
            except InternalError:
                if self.configuration.global_parameters_protocol_level > 0:
                    Miscellaneous.protocol(
                        "Error: '" + name + "' does not contain valid input for a stacking job,"
                                           " continune with next job.\n",
                        self.workflow.attached_log_file)

        self.job_number = len(self.jobs)
        if self.job_number == 0:
            if self.configuration.global_parameters_protocol_level > 0:
                Miscellaneous.protocol(
                    "Error: No valid job specified, execution halted.",
                    self.workflow.attached_log_file)
            self.stop_execution()

        self.job_index = 0

        # If a dark frame was specified, load it.
        if arguments.dark:
            if self.configuration.global_parameters_protocol_level > 0:
                Miscellaneous.protocol("+++ Loading master dark frame +++",
                                       self.workflow.attached_log_file)
            self.signal_load_master_dark.emit(arguments.dark)

        # If a flat frame was specified, load it.
        if arguments.flat:
            if self.configuration.global_parameters_protocol_level > 0:
                Miscellaneous.protocol("+++ Loading master flat frame +++",
                                       self.workflow.attached_log_file)
            self.signal_load_master_flat.emit(arguments.flat)

    @QtCore.pyqtSlot(str)
    def report_calibration_error(self, message):
        if self.configuration.global_parameters_protocol_level > 0:
            Miscellaneous.protocol("           " + message,
                                   self.workflow.attached_log_file, precede_with_timestamp=False)

    @QtCore.pyqtSlot(str)
    def report_error(self, message):
        """
        This method is triggered by the workflow thread via a signal when an error is to be
        reported. Depending on the protocol level, the error message is written to the
        protocol (file).

        :param message: Error message to be displayed
        :return: -
        """

        if self.configuration.global_parameters_protocol_level > 0:
            Miscellaneous.protocol(message + "\n", self.workflow.attached_log_file)

    @QtCore.pyqtSlot(str)
    def next_job_after_error(self, message):
        """
        This method is triggered by the workflow thread via a signal when an error causes a job to
        be aborted. Depending on the protocol level, the error message is written to the
        protocol (file).

        :param message: Error message to be displayed
        :return: -
        """

        # Report the error.
        self.report_error(message)

        # Abort the current job and go to the next one.
        self.work_next_task("Next job")

    def work_next_task(self, next_activity):
        """
        This is the central place where all activities are scheduled. Depending on the
         "next_activity" chosen, the appropriate activity is started on the workflow thread.

        :param next_activity: Activity to be performed next.
        :return: -
        """

        # Make sure not to process an empty job list, or a job index out of range.
        if not self.jobs or self.job_index >= self.job_number:
            return

        self.activity = next_activity

        # Start workflow activities. When a workflow method terminates, it invokes this method on
        # the GUI thread, with "next_activity" denoting the next step in the processing chain.
        if self.activity == "Read frames":

            # For the first activity (reading all frames from the file system) there is no
            # GUI interaction. Start the workflow action immediately.
            self.signal_frames.emit(self.jobs[self.job_index])

        elif self.activity == "Rank frames":

            # In batch mode no frames can be dropped in the user dialog.
            update_index_translation_table = False
            # Now start the corresponding action on the workflow thread.
            self.signal_rank_frames.emit(update_index_translation_table)

        elif self.activity == "Select frames":

            # The dialog to exclude frames is not to be called. Go to frames alignment
            # immediately.
            self.signal_align_frames.emit(0, 0, 0, 0)

        elif self.activity == "Select stack size":

            # In automatic mode, nothing is to be done in the workflow thread. Start the next
            # activity on the main thread immediately.
            self.workflow.work_next_task_signal.emit("Set ROI")

        elif self.activity == "Set ROI":

            # If all index bounds are set to zero, no ROI is selected.
            self.signal_set_roi.emit(0, 0, 0, 0)

        elif self.activity == "Set alignment points":

            # In automatic mode, compute the AP grid automatically in the workflow thread. In this
            # case, the AlignmentPoints object is created there as well.
            self.signal_set_alignment_points.emit()

        elif self.activity == "Compute frame qualities":
            self.signal_compute_frame_qualities.emit()

        elif self.activity == "Stack frames":
            self.signal_stack_frames.emit()

        elif self.activity == "Save stacked image":
            self.signal_save_stacked_image.emit()

        elif self.activity == "Next job":
            self.job_index += 1
            if self.job_index < self.job_number:
                # If the end of the queue is not reached yet, start with reading frames of next job.
                self.activity = "Read frames"
                self.signal_frames.emit(self.jobs[self.job_index])
            else:
                self.stop_execution()

    def print_arguments(self, arguments):
        """
        This is an auxiliary method for debugging. It prints all arguments passed to the program.

        :param arguments: Arguments object created by the ArgumentParser
        :return: -
        """

        print("Jobs: " + str(arguments.job_input))
        print("Store protocol with results: " + str(arguments.protocol))
        print("Protocol detail level: " + str(arguments.protocol_detail))
        print("Buffering level: " + str(arguments.buffering_level))
        print("Maximum RAM size: " + str(arguments.ram_size))
        print("Image format for output: " + arguments.out_format)
        print("Add number of stacked frames to output file name: " + str(arguments.name_add_f))
        print("Add percentage of stacked frames to output file name: " + str(arguments.name_add_p))
        print(
            "Add alignment point box size (pixels) to output file name: " + str(
                arguments.name_add_apb))
        print("Add number of alignment points to output file name: " + str(arguments.name_add_apn))
        print("")
        print("Debayering option: " + arguments.debayering)
        print("Debayering method: " + arguments.debayer_method)
        print("Noise level: " + str(arguments.noise))
        print("Frame stabilization mode: " + arguments.stab_mode)
        print("Stabilization patch size (% of frame): " + str(arguments.stab_size))
        print("Stabilization search width (pixels): " + str(arguments.stab_sw))
        print("Percentage of best frames for reference frame computation: " + str(
            arguments.rf_percent))
        if arguments.dark:
            print("Image file for dark frame correction: " + arguments.dark)
        if arguments.flat:
            print("Image file for flat frame correction: " + arguments.flat)
        print("")
        print("Alignment point box width (pixels): " + str(arguments.align_box_width))
        print("Alignment point search width (pixels): " + str(arguments.align_search_width))
        print("Alignment point minimum structure: " + str(arguments.align_min_struct))
        print("Alignment point minimum brightness: " + str(arguments.align_min_bright))
        print("")
        print("Percentage of best frames to be stacked: " + str(arguments.stack_percent))
        print("Number of best frames to be stacked: " + str(arguments.stack_number))
        print("Normalize frame brightness: " + str(arguments.normalize_bright))
        print("Normalization black cut-off: " + str(arguments.normalize_bco))

    def stop_execution(self):
        """
        Halt the application. (There might be a more elegant way to do this!)

        :return: -
        """
        # Wait a little before exiting, so that output buffers can be purged.
        sleep(0.2)
        quit(0)



# -*- coding: utf-8; -*-
"""
Copyright (c) 2019 Rolf Hempel, rolf6419@gmx.de

This file is part of the PlanetarySystemStacker tool (PSS).
https://github.com/Rolf-Hempel/PlanetarySystemStacker

PSS is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PSS.  If not, see <http://www.gnu.org/licenses/>.

"""

from numpy import arange
from PyQt5 import QtWidgets
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as Canvas
from matplotlib.figure import Figure

from shift_distribution_viewer_gui import Ui_shift_distribution_viewer


class MatplotlibWidget(Canvas):
    """
    This widget creates a plot of frame qualities, either sorted chronologically or by quality.

    """

    def __init__(self, parent=None):
        """
        Initialize the widget.

        :param parent: Parent object
        """

        super(MatplotlibWidget, self).__init__(Figure())

        self.setParent(parent)

        plt.rcParams.update({'font.size': 8})
        self.fig, self.ax = plt.subplots()

    def draw_distribution(self, shift_distribution):
        """
        Draw the shift distribution.

        :param shift_distribution: 1D Numpy array (int) with counts for each shift size.
        :return: -
        """

        # Find the last non-zero entry in the array. If no alignment points have been set,
        # show a message in the widget title.
        if max(shift_distribution) > 0:
            max_index = [index for index, item in enumerate(shift_distribution) if item != 0][-1] + 1
            plt.title('Frequency distribution of local warp sizes at alignment points')
        else:
            max_index = 1
            plt.title('No alignment points, therefore no warp distribution')

        pixels = arange(max_index)

        plt.bar(pixels, shift_distribution[:max_index], align='center')
        plt.xticks(pixels, pixels)
        plt.xlabel('Warp size (pixels)')
        plt.ylabel('Frequency')
        self.fig.canvas.draw()
        self.fig.canvas.flush_events()


class ShiftDistributionViewerWidget(QtWidgets.QFrame, Ui_shift_distribution_viewer):
    """
    Display the distribution of frame shifts at alignment points. For each AP in each contributing
    frame the total shift (in pixels) is rounded to the next integer value. For every integer number
    up to the maximum frame shift the number of occurrences is displayed.
    """

    def __init__(self, parent_gui, shift_distribution, shift_failure_percent, signal_finished,
                 parent=None):
        """
        Initialize the Viewer. The widget has a fixed size and is rendered as a QFrame.

        :param parent_gui: GUI object by which the viewer is invoked.
        :param parent: Parent object
        :param shift_distribution: 1D Numpy array (int) with counts for each shift size.
        :param shift_failure_percent: Percent of failed warp shift measurements.
        :param signal_finished: Qt signal with signature () telling the workflow thread that the
                                viewer has finished, or None (no signalling).
        """

        QtWidgets.QFrame.__init__(self, parent)
        self.setupUi(self)

        self.setFrameShape(QtWidgets.QFrame.Panel)
        self.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.setObjectName("shift_distribution_viewer")

        self.setFixedSize(900, 600)

        self.parent_gui = parent_gui
        self.shift_distribution = shift_distribution
        self.signal_finished = signal_finished

        self.buttonBox.accepted.connect(self.accept)

        # Create the Matplotlib widget showing the shift statistics.
        self.matplotlib_widget = MatplotlibWidget()
        self.verticalLayout.insertWidget(0, Canvas(self.matplotlib_widget.fig))

        self.matplotlib_widget.draw_distribution(self.shift_distribution)

        # Display the fraction of failed shift measurements. Use red ink if the value exceeds 5%.
        if shift_failure_percent >= 0.:
            self.failedShiftsLabel.setText(
                "Failed shift measurements: " + str(shift_failure_percent) + "%")
            if shift_failure_percent > 5.:
                self.failedShiftsLabel.setStyleSheet('color: red')
            else:
                self.failedShiftsLabel.setStyleSheet('color: black')
        else:
            self.failedShiftsLabel.setText("")

    def accept(self):
        """

        :return: -
        """

        # If a signal was passed in initialization, tell the workflow thread that the viewer has
        # finished.
        if self.signal_finished is not None:
            self.signal_finished.emit()

        self.close()

    def closeEvent(self, event):
        """
        Remove the widget from the parent GUI and close the viewer.

        :param event: Close event object
        :return: -
        """

        # Remove the viewer widget from the main GUI and exit.
        self.parent_gui.display_widget(None, display=False)
        self.close()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'version_manager_widget.ui'
#
# Created by: PyQt5 UI code generator 5.11.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_version_manager_widget(object):
    def setupUi(self, version_manager_widget):
        version_manager_widget.setObjectName("version_manager_widget")
        version_manager_widget.resize(401, 104)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(version_manager_widget.sizePolicy().hasHeightForWidth())
        version_manager_widget.setSizePolicy(sizePolicy)
        self.verticalLayout = QtWidgets.QVBoxLayout(version_manager_widget)
        self.verticalLayout.setObjectName("verticalLayout")
        self.groupBox = QtWidgets.QGroupBox(version_manager_widget)
        self.groupBox.setObjectName("groupBox")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.groupBox)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.pushButton_new = QtWidgets.QPushButton(self.groupBox)
        self.pushButton_new.setObjectName("pushButton_new")
        self.gridLayout_2.addWidget(self.pushButton_new, 1, 0, 1, 2)
        self.pushButton_delete = QtWidgets.QPushButton(self.groupBox)
        self.pushButton_delete.setObjectName("pushButton_delete")
        self.gridLayout_2.addWidget(self.pushButton_delete, 1, 2, 1, 2)
        self.pushButton_save = QtWidgets.QPushButton(self.groupBox)
        self.pushButton_save.setObjectName("pushButton_save")
        self.gridLayout_2.addWidget(self.pushButton_save, 1, 4, 1, 2)
        self.pushButton_save_as = QtWidgets.QPushButton(self.groupBox)
        self.pushButton_save_as.setObjectName("pushButton_save_as")
        self.gridLayout_2.addWidget(self.pushButton_save_as, 1, 6, 1, 2)
        self.spinBox_compare = QtWidgets.QSpinBox(self.groupBox)
        self.spinBox_compare.setObjectName("spinBox_compare")
        self.gridLayout_2.addWidget(self.spinBox_compare, 0, 7, 1, 1)
        self.checkBox_blink_compare = QtWidgets.QCheckBox(self.groupBox)
        self.checkBox_blink_compare.setObjectName("checkBox_blink_compare")
        self.gridLayout_2.addWidget(self.checkBox_blink_compare, 0, 4, 1, 3)
        self.spinBox_version = QtWidgets.QSpinBox(self.groupBox)
        self.spinBox_version.setObjectName("spinBox_version")
        self.gridLayout_2.addWidget(self.spinBox_version, 0, 2, 1, 1)
        self.label_version = QtWidgets.QLabel(self.groupBox)
        self.label_version.setObjectName("label_version")
        self.gridLayout_2.addWidget(self.label_version, 0, 0, 1, 2)
        self.verticalLayout.addWidget(self.groupBox)

        self.retranslateUi(version_manager_widget)
        QtCore.QMetaObject.connectSlotsByName(version_manager_widget)

    def retranslateUi(self, version_manager_widget):
        _translate = QtCore.QCoreApplication.translate
        version_manager_widget.setWindowTitle(_translate("version_manager_widget", "Form"))
        version_manager_widget.setToolTip(_translate("version_manager_widget", "Create a new postprocessing version"))
        self.groupBox.setTitle(_translate("version_manager_widget", "Version manager"))
        self.pushButton_new.setToolTip(_translate("version_manager_widget", "Create a new postprocessing version"))
        self.pushButton_new.setText(_translate("version_manager_widget", "New"))
        self.pushButton_delete.setToolTip(_translate("version_manager_widget", "Delete the selected postprocessing version"))
        self.pushButton_delete.setText(_translate("version_manager_widget", "Delete"))
        self.pushButton_save.setToolTip(_translate("version_manager_widget", "Save the selected postprocessing version using the standard suffix"))
        self.pushButton_save.setText(_translate("version_manager_widget", "Save"))
        self.pushButton_save_as.setToolTip(_translate("version_manager_widget", "Save the selected postprocessing version at a location of choice"))
        self.pushButton_save_as.setText(_translate("version_manager_widget", "Save as"))
        self.spinBox_compare.setToolTip(_translate("version_manager_widget", "Select a version to compare the selected one with"))
        self.checkBox_blink_compare.setToolTip(_translate("version_manager_widget", "Alternate between selected and some other version"))
        self.checkBox_blink_compare.setText(_translate("version_manager_widget", "Blink compare with"))
        self.spinBox_version.setToolTip(_translate("version_manager_widget", "Select a postprocessing version. The unprocessed image is version \"0\"."))
        self.label_version.setText(_translate("version_manager_widget", "Version selected"))




# -*- coding: utf-8 -*-
"""
Created on Fri Sep 21 15:39:35 2018

@author: Michal Powalko m.powalko@gmail.com
"""

import time
import numpy as np


# Currently by Rolf
tic = time.time()
buffer_Rolf = np.empty([200, 1920, 1080])
mem_alloc_Rolf = buffer_Rolf.nbytes / 1024**3
print('Preallocated {0:.3f} [GB] memory for importing the data in {1:.3f} [s]...Rolf'.format(mem_alloc_Rolf, time.time() - tic))

# Proposal from Michal
# "buffer" is a matrix to HOLD "input data" - and this consist of 8bit data = "uint8". Specifing the "dtype"
# as "np.uint8" You can save al lot of mememory and gain speed (in case of many frames)
# The results of "mean/median" will be saved under another variable, where the resolution might be adjustet
# The only disadvantageous would be, when working with input files with greater color resolution than 8bit

tic = time.time()
buffer_Michal = np.empty([200, 1920, 1080], dtype = np.uint8)
mem_alloc_Michal = buffer_Michal.nbytes / 1024**3
print('Preallocated {0:.3f} [GB] memory for importing the data in {1:.3f} [s]...Michal :)'.format(mem_alloc_Michal, time.time() - tic))

#------------------------------------------------------------------------------------------------------------------------

# Currently by Rolf
tic = time.time()
mean_frame_Rolf = np.mean(buffer_Rolf, axis=0)
print('Calculated "mean" from imported frames in {0:.3f} [s]...Rolf'.format(time.time() - tic))


# Proposal from Michal
# Taking advantage of "variable.npCommand" might save some time
# Disadvantage may be luck of some internal check in numpy "sanity check of input data", but I have never had a problem with this
buffer_Rolf2 = np.empty([200, 1920, 1080])
tic = time.time()
mean_frame_Michal = buffer_Rolf2.mean(axis=0)
print('Calculated "mean" from imported frames in {0:.3f} [s]...Michal'.format(time.time() - tic))

#------------------------------------------------------------------------------------------------------------------------

# Proposal from Michal
# Calculating Sum as well
# Having read:   http://www.stark-labs.com/craig/resources/Articles-&-Reviews/BitDepthStacking.pdf
# building a sum from 8bits we can gaint the bit resolution. To prevent the resolution it is necessary to either:
# - saving stacked uint8 frames with "float"resolution (to keep the "after comma values)
# - or to sum to uint>8, like uint16 or even uint32
tic = time.time()
sum_frame = np.sum(buffer_Rolf, axis=0, dtype=np.uint32)
print('Calculated "sum" from imported frames in {0:.3f} [s]...'.format(time.time() - tic))

#------------------------------------------------------------------------------------------------------------------------

# Proposal from Michal
# Calculating Median as well

tic = time.time()
median_frame = np.median(buffer_Rolf, axis=0)
print('Calculated "median" from imported frames in {0:.3f} [s]...'.format(time.time() - tic))

# We can boost performance by setting "overwrite_input=True" --> the original matrix will be "reshaped" instead of creating another one
# "variable.median" is not available :(
# https://docs.scipy.org/doc/numpy/reference/generated/numpy.median.html
buffer_Rolf2 = np.empty([200, 1920, 1080])
tic = time.time()
median_frame2 = np.median(buffer_Rolf2, axis=0, overwrite_input=True)
print('Calculated "median" from imported frames in {0:.3f} [s]...with "overwrite_input=True"'.format(time.time() - tic))


#------------------------------------------------------------------------------------------------------------------------
## TIPS

# setting "dtype=np.uint32" inside numpy function is much faster then "().astype(np.uint32)"

# "variable.npCommand" is slightly faster than "npCommand(variable)"

# Downsampling to 8bit
# This may be tricky --> ".astype(np.uint)" will perform "floor" instead of "round".
# This might lead to situation, where "9,9999" will be "rounded" to "9"
downsample_ipnut = 10 * np.random.rand(3,2)
downsample_wrong = downsample_ipnut.astype(np.uint8)
downsample_good = downsample_ipnut.round().astype(np.uint8)



from cv2 import imread, IMREAD_UNCHANGED, imshow, waitKey, normalize, NORM_MINMAX
from numpy import uint8

# filename = 'Photodisc.png'
# filename = "E:\SW-Development\Python\PlanetarySystemStacker\planetary_system_stacker\Images\\another_short_video.jpg"
filename = "E:\SW-Development\Python\PlanetarySystemStacker\Examples\AVI_Chris-Garry\\20110929_005012_jupiter_gbrg_ULRA_pss_gpp.jpg"
input_image = imread(filename, IMREAD_UNCHANGED)
display_image = normalize(input_image, None, alpha=0, beta=255, norm_type=NORM_MINMAX)
imshow('image',display_image)
waitKey(0)



import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from cv2 import VideoCapture, VideoWriter_fourcc, VideoWriter
from numpy import full, zeros, uint8, clip


def create_dark(dark):
    dark[700:800, 700:800, :] += 15
    return dark


def create_flat(image):
    type = image.dtype
    image[100:200, 100:200, :] = (image[100:200, 100:200, :] * 0.9).astype(type)
    return image


original_video = 'Videos/short_video.avi'
processed_video = 'D:/SW-Development/Python/PlanetarySystemStacker/Examples/Darks_and_Flats/short_video_dark-flat-processed.avi'
video_dark = "D:/SW-Development/Python/PlanetarySystemStacker/Examples/Darks_and_Flats/artificial_dark.avi"
video_flat = "D:/SW-Development/Python/PlanetarySystemStacker/Examples/Darks_and_Flats/artificial_flat.avi"
ny = 960
nx = 1280

dark = create_dark(zeros((ny, nx, 3), uint8))
flat = (create_flat(full((ny, nx, 3), 128)) + dark).astype(uint8)

plt.imshow(dark)
plt.show()
plt.imshow(flat)
plt.show()

# Create the VideoCapture object.
cap = VideoCapture(original_video)
# Define the codec and create VideoWriter object
fourcc = VideoWriter_fourcc(*'XVID')
out = VideoWriter(processed_video, fourcc, 20.0, (nx, ny))
out_dark = VideoWriter(video_dark, fourcc, 20.0, (nx, ny))
out_flat = VideoWriter(video_flat, fourcc, 20.0, (nx, ny))

while (True):
    # Capture frame-by-frame
    ret, frame = cap.read()
    if ret is True:
        frame_flat_applied = create_flat(frame)
        frame_dark_added = frame_flat_applied + dark
        frame_processed = clip(frame_dark_added, 0, 255)
        out.write(frame_processed)
        out_dark.write(dark)
        out_flat.write(flat)
    else:
        break

# When everything done, release the capture
cap.release()
out.release()
out_dark.release()
out_flat.release()



# -*- coding: utf-8; -*-
"""
Copyright (c) 2018 Rolf Hempel, rolf6419@gmx.de

This file is part of the PlanetarySystemStacker tool (PSS).
https://github.com/Rolf-Hempel/PlanetarySystemStacker

PSS is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PSS.  If not, see <http://www.gnu.org/licenses/>.

Part of this module (in class "AlignmentPointEditor" was copied from
https://stackoverflow.com/questions/35508711/how-to-enable-pan-and-zoom-in-a-qgraphicsview

"""
from glob import glob
from sys import argv, exit
from time import time

from PyQt5 import QtCore, QtGui, QtWidgets
from numpy import uint8

from align_frames import AlignFrames
from alignment_point_editor_gui import Ui_alignment_point_editor
from alignment_points import AlignmentPoints
from configuration import Configuration, ConfigurationParameters
from exceptions import InternalError, NotSupportedError, Error
from frame_viewer import FrameViewer
from frames import Frames
from rank_frames import RankFrames


class GraphicsScene(QtWidgets.QGraphicsScene):
    """
    Handle mouse events on the graphics scene depicting the image with the alignment points (APs)
    superimposed. The left mouse button can be used to create APs, to move them, and to change their
    sizes. Using the right mouse button APs can be deleted.

    """

    def __init__(self, photo_editor, parent=None):
        """
        Initialize the scene object.

        :param photo_editor: the object in which the scene is defined
        :param parent: The parent class
        """

        QtWidgets.QGraphicsScene.__init__(self, parent)
        self.photo_editor = photo_editor
        self.left_button_pressed = False
        self.right_button_pressed = False
        self.moved_ap = None
        self.new_ap = False
        self.remember_ap = None

        # Set the maximum diestance between a right click and a right release up to which they
        # are identified with each other. If the distance is larger, the mouse events are
        # identified with the opening of a rectangular patch between.
        self.single_click_threshold = 5
        # Set the maximum distance up to which a mouse position is identified with an existing
        # alignment point location.
        self.max_match_distance = 5
        # Set the factor by which the AP size is changed with a single mouse wheel stop.
        self.ap_size_change_factor = 1.05

    def mousePressEvent(self, event):
        """
        A nouse button is pressed. This can either be a lft or a right button.

        :param event: event object
        :return: -
        """

        # The following actions are not performed in drag-and-zoom mode. The switch between both
        # modes is handled in the higher-level object "photo_viewer".
        if not self.photo_editor.drag_mode:
            pos = event.lastScenePos()
            x = int(pos.x())
            y = int(pos.y())

            # The left button is pressed.
            if event.button() == QtCore.Qt.LeftButton:
                self.left_button_pressed = True

                # Find the closest AP.
                neighbor_ap, distance = self.photo_editor.aps.find_neighbor(y, x,
                                        self.photo_editor.aps.alignment_points)

                # If the AP list is not empty and the closest distance is very small, assume that
                # the AP is to be moved.
                if neighbor_ap and distance < self.max_match_distance:
                    self.moved_ap = neighbor_ap
                    self.remember_ap = neighbor_ap.copy()

                # Create a new AP.
                else:
                    # Create a preliminary AP with the computed size. It only becomes a real AP when
                    # the mouse is released.
                    self.remember_ap = self.photo_editor.aps.new_alignment_point(y, x, False, False,
                        False, False)
                    if self.remember_ap:
                        self.new_ap = True

            # The right button is pressed.
            elif event.button() == QtCore.Qt.RightButton:
                self.right_button_pressed = True

                # Remember the location and initialize an object which during mouse moving stores
                # the rectangular patch opening between start and end positions.
                self.right_y_start = pos.y()
                self.right_x_start = pos.x()
                self.remember_sr = None

    def mouseReleaseEvent(self, event):
        """
        A mouse button is released.

        :param event: event object
        :return: -
        """

        if not self.photo_editor.drag_mode:
            pos = event.lastScenePos()
            x = int(pos.x())
            y = int(pos.y())

            # The left button is released.
            if event.button() == QtCore.Qt.LeftButton:
                self.left_button_pressed = False

                # An existing AP was moved, replace it with the moved one.
                if self.moved_ap:
                    self.removeItem(self.remember_ap['graphics_item'])
                    self.photo_editor.replace_alignment_point(self.moved_ap,
                                                              self.remember_ap)
                    self.moved_ap = None
                # A new AP was created, and possibly moved. Add it to the list.
                elif self.new_ap:
                    self.removeItem(self.remember_ap['graphics_item'])
                    self.photo_editor.add_alignment_point(self.remember_ap)
                    self.new_ap = False

            # The right button is released.
            elif event.button() == QtCore.Qt.RightButton:
                self.right_button_pressed = False

                # If the mouse was not moved much between press and release, a single AP is deleted.
                if max(abs(y - self.right_y_start),
                       abs(x - self.right_x_start)) < self.single_click_threshold:

                    # Find the closest AP and remove it from the scene and the AP list.
                    ap, dist = self.photo_editor.aps.find_neighbor(y, x,
                               self.photo_editor.aps.alignment_points)
                    self.photo_editor.remove_alignment_points([ap])

                # The mouse was moved between press and release. Remove all APs in the opening
                # rectangular patch, both from the scene and the AP list.
                else:
                    y_low = min(self.right_y_start, self.right_y_end)
                    y_high = max(self.right_y_start, self.right_y_end)
                    x_low = min(self.right_x_start, self.right_x_end)
                    x_high = max(self.right_x_start, self.right_x_end)
                    remove_ap_list = self.photo_editor.aps.find_alignment_points(y_low, y_high,
                                                                                 x_low, x_high)
                    self.removeItem(self.remember_sr)
                    self.photo_editor.remove_alignment_points(remove_ap_list)
            self.update()

    def mouseMoveEvent(self, event):
        """
        The mouse is moved while being pressed.

        :param event: event object
        :return: -
        """

        pos = event.lastScenePos()
        self.x = int(pos.x())
        self.y = int(pos.y())

        if not self.photo_editor.drag_mode:
            # The mouse is moved with the left button pressed.
            if self.left_button_pressed and (self.moved_ap or self.new_ap):

                # Copy the preliminary AP, and apply the changes to the copy only.
                new_ap = self.remember_ap.copy()

                # Move the preliminary AP to the new coordinates.
                new_ap = self.photo_editor.aps.move_alignment_point(new_ap, self.y, self.x)
                # The move was successful. Replace the preliminary AP with the new one.
                if new_ap:
                    # The scene widget corresponding to the preliminary AP was stored with the AP.
                    # Remove it from the scene before the moved AP is drawn.
                    self.removeItem(self.remember_ap['graphics_item'])
                    # Draw the new preliminary AP.
                    new_ap['graphics_item'] = AlignmentPointGraphicsItem(new_ap)
                    self.addItem(new_ap['graphics_item'])
                    self.remember_ap = new_ap

            # The mouse is moved with the right button pressed.
            elif self.right_button_pressed:
                self.right_y_end = self.y
                self.right_x_end = self.x

                # Compute the new rectangle for selecting APs to be removed.
                new_sr = SelectionRectangleGraphicsItem(self.right_y_start, self.right_x_start,
                    self.y, self.x)

                # If the rectangle was drawn for a previous location, replace it with the new one.
                if self.remember_sr is not None:
                    self.removeItem(self.remember_sr)
                self.addItem(new_sr)
                self.remember_sr = new_sr
            self.update()

    def wheelEvent(self, event):
        """
        Handle scroll events. Change the size of the nearest AP.

        :param event: wheel event object
        :return: -
        """

        # Depending of wheel direction, set the factor to greater or smaller than 1.
        self.change_ap_size(event.angleDelta().y())

    def keyPressEvent(self, event):
        """
        Handle key events as an alternative to wheel events. Change the size of the nearest AP.

        :param event: key event
        :return: -
        """

        # This is a workaround: Instead of "93" it should read "QtCore.Qt.Key_Plus", but that
        # returns 43 instead.
        if event.key() == 93 and event.modifiers() & QtCore.Qt.ControlModifier:
            self.change_ap_size(1)
        elif event.key() == QtCore.Qt.Key_Minus and event.modifiers() & QtCore.Qt.ControlModifier:
            self.change_ap_size(-1)

    def change_ap_size(self, direction):
        """
        Change the size of the nearest AP.

        :param direction: If > 0, increase the size by a fixed factor. If < 0, decrease its size.
        :return: -
        """

        # Find the closest AP.
        ap, dist = self.photo_editor.aps.find_neighbor(self.y, self.x,
                                                       self.photo_editor.aps.alignment_points)

        # If no AP has been set, there is nothing do be done.
        if ap is None:
            return

        # Depending of direction value, set the factor to greater or smaller than 1.
        if direction > 0:
            factor = self.ap_size_change_factor
        else:
            factor = 1. / self.ap_size_change_factor

        # Copy the AP, and apply the changes to the copy only.
        new_ap = ap.copy()
        if self.photo_editor.aps.resize_alignment_point(new_ap, factor):
            new_ap['graphics_item'] = AlignmentPointGraphicsItem(new_ap)
            # Replace the old AP with the resized version of it.
            self.photo_editor.replace_alignment_point(ap, new_ap)


class AlignmentPointGraphicsItem(QtWidgets.QGraphicsItem):
    """
    This widget represents an AP. It consists of a small red dot at the AP location, a red
    bounding rectangle, and a transparent green filling.
    """

    def __init__(self, ap):
        super(AlignmentPointGraphicsItem, self).__init__()

        # Set the color and transparency of the filling.
        self.color_surface = QtGui.QColor(255, 255, 0, 60)

        # Set the color of the bouding rectangle and central dot.
        self.color_boundary = QtGui.QColor(255, 0, 0)
        self.y = ap["y"]
        self.x = ap["x"]
        self.patch_y_low = ap["patch_y_low"]
        self.patch_y_high = ap["patch_y_high"]
        self.patch_x_low = ap["patch_x_low"]
        self.patch_x_high = ap["patch_x_high"]
        self.pen_boundary = QtGui.QPen(self.color_boundary)
        self.pen_boundary.setStyle(0)
        self.width_x = self.patch_x_high - self.patch_x_low
        self.width_x_external = self.width_x + self.pen_boundary.width()
        self.width_y = self.patch_y_high - self.patch_y_low
        self.width_y_external = self.width_y + self.pen_boundary.width()

        # Set the size of the central dot.
        self.dot_width = max(1, int(self.width_x / 30))

    def boundingRect(self):
        return QtCore.QRectF(self.patch_x_low, self.patch_y_low, self.width_x_external,
                             self.width_y_external)

    def paint(self, painter, option, widget):
        painter.setPen(self.pen_boundary)
        painter.setBrush(self.color_boundary)
        painter.drawEllipse(self.x, self.y, self.dot_width, self.dot_width)
        painter.setBrush(self.color_surface)
        painter.drawRect(self.patch_x_low, self.patch_y_low, self.width_x, self.width_y)


class SelectionRectangleGraphicsItem(QtWidgets.QGraphicsItem):
    """
    This widget represents the selection rectangle opening when the mouse is being moved with the
    right mouse button  pressed.
    """

    def __init__(self, y_start, x_start, y_end, x_end):
        super(SelectionRectangleGraphicsItem, self).__init__()

        # Set the color of the transparent rectangle filling.
        self.color_surface = QtGui.QColor(0, 0, 255, 40)

        # Set the color of the bounding rectangle.
        self.color_boundary = QtGui.QColor(255, 0, 0)

        # Set the coordinate limits of the rectangle. Start and end locations can be anywhere.
        self.y_low = min(y_start, y_end)
        self.x_low = min(x_start, x_end)
        self.y_high = max(y_start, y_end)
        self.x_high = max(x_start, x_end)
        self.pen_boundary = QtGui.QPen(self.color_boundary)
        self.pen_boundary.setWidth(1)
        self.width_y = self.y_high - self.y_low
        self.width_x = self.x_high - self.x_low
        self.width_y_external = self.width_y + self.pen_boundary.width()
        self.width_x_external = self.width_x + self.pen_boundary.width()

    def boundingRect(self):
        return QtCore.QRectF(self.x_low, self.y_low, self.width_x_external, self.width_y_external)

    def paint(self, painter, option, widget):
        painter.setPen(self.pen_boundary)
        painter.setBrush(self.color_surface)
        painter.drawRect(self.x_low, self.y_low, self.width_x, self.width_y)


class AlignmentPointEditor(FrameViewer):
    """
    This widget implements an editor for handling APs superimposed onto an image. It supports two
    modes:
    - In "drag mode" the mouse can be used for panning, and the scroll wheel for zooming.
    - In "alignment point mode" the mouse is used to create/remove APs, to move them or to change
      their sizes.
    The "cntrl" key is used to switch between the two modes.
    """

    def __init__(self, image, alignment_points):
        super(AlignmentPointEditor, self).__init__()
        self.image = image

        # Initialize the undo stack.
        self.undoStack = QtWidgets.QUndoStack(self)

        # Initialize the alignment point object.
        self.aps = alignment_points

        # Load the image.
        self.setPhoto(self.image)

    def initialize_scene(self):
        """
        Initialize the scene. This object handles mouse events if not in drag mode. In derived
        viewer classes this method is replaced with the instantiation of a custom version of the
        graphics scene.

        :return:
        """

        self._scene = GraphicsScene(self, self)
        self._scene.addItem(self._photo)
        self.setScene(self._scene)

    def keyPressEvent(self, event):
        """
        The control key is used to switch between drag and AP modes.

        :param event: event object
        :return: -
        """

        # If the control key is pressed, switch to "no drag mode".
        if event.key() == QtCore.Qt.Key_Z and event.modifiers() & QtCore.Qt.ControlModifier:
            self.undoStack.undo()
        elif event.key() == QtCore.Qt.Key_Y and event.modifiers() & QtCore.Qt.ControlModifier:
            self.undoStack.redo()
        else:
            super(AlignmentPointEditor, self).keyPressEvent(event)

    def createApGrid(self):
        """
        Create an initial alignment point grid, and display it in the viewer.

        :return: -
        """

        command = CommandCreateApGrid(self)
        self.undoStack.push(command)

    def add_alignment_point(self, ap):
        """
        Add an AP using the undo framework.

        :param ap: AP object in the AP list
        :return: -
        """

        command = CommandAdd(self, self.aps, ap)
        self.undoStack.push(command)

    def remove_alignment_points(self, ap_list):
        """
        Remove an AP using the undo framework.

        :param ap: AP object in the AP list
        :return: -
        """

        command = CommandRemove(self, self.aps, ap_list)
        self.undoStack.push(command)

    def replace_alignment_point(self, ap_old, ap_new):
        """
        Replace an AP with another one using the undo framework.

        :param ap_old: AP object in the AP list to be replaced
        :param ap_new: new AP object to be added to the AP list
        :return: -
        """

        command = CommandReplace(self, self.aps, ap_old, ap_new)
        self.undoStack.push(command)


class CommandCreateApGrid(QtWidgets.QUndoCommand):
    """
    Undoable command to replace all existing APs with a new AP grid.
    """
    def __init__(self, photo_editor):
        super(CommandCreateApGrid, self).__init__()
        self.photo_editor = photo_editor
        # Copy the old ap list before overwriting it with the new AP grid.
        self.old_ap_list = self.photo_editor.aps.alignment_points
        # Create the new AP grid.
        self.photo_editor.aps.create_ap_grid()
        # Copy the new list of APs.
        self.new_ap_list = self.photo_editor.aps.alignment_points
        for ap in self.new_ap_list:
            ap['graphics_item'] = AlignmentPointGraphicsItem(ap)

    def redo(self):
        # Remove all APs from the scene.
        for item in self.photo_editor._scene.items():
            if isinstance(item, AlignmentPointGraphicsItem):
                self.photo_editor._scene.removeItem(item)

        # Copy the AP list from grid generation to the "alignment_point" object.
        self.photo_editor.aps.alignment_points = self.new_ap_list.copy()
        # Draw all new APs and update the scene.
        for ap in self.photo_editor.aps.alignment_points:
            self.photo_editor._scene.addItem(ap['graphics_item'])
        self.photo_editor._scene.update()
        self.photo_editor.setFocus()

    def undo(self):
        # Restore the old AP list.
        self.photo_editor.aps.alignment_points = self.old_ap_list

        # Remove all durrent APs from the scene.
        for item in self.photo_editor._scene.items():
            if isinstance(item, AlignmentPointGraphicsItem):
                self.photo_editor._scene.removeItem(item)
        # Draw all old APs and update the scene.
        for ap in self.photo_editor.aps.alignment_points:
            self.photo_editor._scene.addItem(ap['graphics_item'])
        self.photo_editor._scene.update()
        self.photo_editor.setFocus()


class CommandAdd(QtWidgets.QUndoCommand):
    """
    Undoable command to add an AP to the AP list.
    """
    def __init__(self, photo_editor, aps_object, ap):
        super(CommandAdd, self).__init__()
        self.photo_editor = photo_editor
        self.aps = aps_object
        self.ap = ap
        self.ap['graphics_item'] = AlignmentPointGraphicsItem(ap)

    def redo(self):
        self.aps.add_alignment_point(self.ap)
        self.photo_editor._scene.addItem(self.ap['graphics_item'])
        self.photo_editor._scene.update()
        self.photo_editor.setFocus()

    def undo(self):
        self.photo_editor._scene.removeItem(self.ap['graphics_item'])
        self.aps.remove_alignment_points([self.ap])
        self.photo_editor._scene.update()
        self.photo_editor.setFocus()


class CommandRemove(QtWidgets.QUndoCommand):
    """
    Undoable command to remove an AP from the AP list.
    """
    def __init__(self, photo_editor, aps_object, ap_list):
        super(CommandRemove, self).__init__()
        self.photo_editor = photo_editor
        self.aps = aps_object
        self.ap_list = ap_list

    def redo(self):
        for ap in self.ap_list:
            self.photo_editor._scene.removeItem(ap['graphics_item'])
        self.aps.remove_alignment_points(self.ap_list)
        self.photo_editor._scene.update()
        self.photo_editor.setFocus()

    def undo(self):
        for ap in self.ap_list:
            self.aps.add_alignment_point(ap)
            self.photo_editor._scene.addItem(ap['graphics_item'])
        self.photo_editor._scene.update()
        self.photo_editor.setFocus()


class CommandReplace(QtWidgets.QUndoCommand):
    """
    Undoable command to replace an AP on the AP list with another one.
    """
    def __init__(self, photo_editor, aps_object, ap_old, ap_new):
        super(CommandReplace, self).__init__()
        self.photo_editor = photo_editor
        self.aps = aps_object
        self.ap_old = ap_old
        self.ap_new = ap_new

    def redo(self):
        self.photo_editor._scene.removeItem(self.ap_old['graphics_item'])
        self.aps.replace_alignment_point(self.ap_old, self.ap_new)
        self.photo_editor._scene.addItem(self.ap_new['graphics_item'])
        self.photo_editor._scene.update()
        self.photo_editor.setFocus()

    def undo(self):
        self.photo_editor._scene.removeItem(self.ap_new['graphics_item'])
        self.aps.replace_alignment_point(self.ap_new, self.ap_old)
        self.photo_editor._scene.addItem(self.ap_old['graphics_item'])
        self.photo_editor._scene.update()
        self.photo_editor.setFocus()


class AlignmentPointEditorWidget(QtWidgets.QFrame, Ui_alignment_point_editor):
    """
    This widget implements the AP viewer, to be used as part of the application GUI.
    """
    def __init__(self, parent_gui, configuration, align_frames, alignment_points, signal_finished,
                 parent=None):
        """
        Initialization of the widget.

        :param parent_gui: Parent GUI object
        :param image: Background image on which the APs are superimposed. Usually, the mean frame is
                      used for this purpose.
        :param configuration: Configuration object with parameters
        :param align_frames: AlignFrames object with global shift information for all frames
        :param alignment_points: Alignment point object
        :param signal_finished: Qt signal telling the workflow thread that the APs have been
                                created successfully
        """

        # QtWidgets.QFrame.__init__(self, parent)
        super(AlignmentPointEditorWidget, self).__init__()
        self.setupUi(self)

        self.parent_gui = parent_gui

        # If the mean frame type is not uint8, values correspond to 16bit resolution.
        if align_frames.mean_frame.dtype != uint8:
            self.mean_frame = align_frames.mean_frame / 256
        else:
            self.mean_frame = align_frames.mean_frame

        self.configuration = configuration
        self.aps = alignment_points
        self.signal_finished = signal_finished

        # Create the viewer frame and insert it into the window.
        self.viewer = AlignmentPointEditor(self.mean_frame, self.aps)
        self.horizontalLayout_2.insertWidget(0, self.viewer)
        self.horizontalLayout_2.setStretch(1, 0)

        # Initialize sliders and their value labels.
        self.initialize_widgets_and_local_parameters(
            self.configuration.alignment_points_half_box_width,
            self.configuration.alignment_points_structure_threshold,
            self.configuration.alignment_points_brightness_threshold)

        # Connect events with activities.
        self.aphbw_slider_value.valueChanged['int'].connect(self.aphbw_changed)
        self.apst_slider_value.valueChanged['int'].connect(self.apst_changed)
        self.apbt_slider_value.valueChanged['int'].connect(self.apbt_changed)
        self.restore_standard_values.clicked.connect(self.restore_standard_parameters)

        self.btnApGrid.clicked.connect(self.viewer.createApGrid)
        self.btnUndo.clicked.connect(self.viewer.undoStack.undo)
        self.btnRedo.clicked.connect(self.viewer.undoStack.redo)
        self.buttonBox.accepted.connect(self.done)
        self.shape_y = None
        self.shape_x = None

    def initialize_widgets_and_local_parameters(self, half_box_width, structure_threshold,
                                                brightness_threshold):
        """
        Initialize GUI widgets with current configuration parameter values.

        :param half_box_width: Half the width of a standard alignment box.
        :param structure_threshold: Minimum structure value for an alignment point
                                    (between 0. and 1.)
        :param brightness_threshold: The brightest pixel must be brighter than this value
                                     (0 < value <256)
        :return: -
        """

        self.aphbw_slider_value.setValue(half_box_width * 2)
        self.aphbw_label_display.setText(str(half_box_width * 2))
        self.apst_slider_value.setValue(round(structure_threshold * 100))
        self.apst_label_display.setText(str(structure_threshold))
        self.apbt_slider_value.setValue(brightness_threshold)
        self.apbt_label_display.setText(str(brightness_threshold))
        self.label_message.setStyleSheet('color: red')


    def aphbw_changed(self, value):
        self.configuration.alignment_points_half_box_width = int(value / 2)
        self.configuration.set_derived_parameters()

    def apst_changed(self, value):
        self.configuration.alignment_points_structure_threshold = value / 100.
        self.apst_label_display.setText(str(self.configuration.alignment_points_structure_threshold))

    def apbt_changed(self, value):
        self.configuration.alignment_points_brightness_threshold = value

    def restore_standard_parameters(self):
        """
        Reset configuration parameters and GUI widget settings to standard values.

        :return: -
        """

        # Create a ConfigurationParameters object with standard values for the three AP parameters.
        config_parameters = ConfigurationParameters()
        config_parameters.set_defaults_ap_editing()

        # Reset configuration parameters to standard values.
        self.configuration.alignment_points_half_box_width = \
            config_parameters.alignment_points_half_box_width
        self.configuration.alignment_points_structure_threshold = \
            config_parameters.alignment_points_structure_threshold
        self.configuration.alignment_points_brightness_threshold = \
            config_parameters.alignment_points_brightness_threshold

        # Initialize sliders and their value labels.
        self.initialize_widgets_and_local_parameters(
            self.configuration.alignment_points_half_box_width,
            self.configuration.alignment_points_structure_threshold,
            self.configuration.alignment_points_brightness_threshold)

    def done(self):
        # On exit from the alignment point editor, allocate buffers for APs which have been
        # changed during editing.
        for ap in self.aps.alignment_points:
            if ap['reference_box'] is None:
                self.aps.set_reference_box(ap, self.aps.mean_frame)

        # Reset the busy status of the higher-level GUI (to enable the "play" button) and
        # update the status of the higher-level GUI (if a reference to it was passed in "init").
        if self.parent_gui is not None:
            self.parent_gui.busy = False
            self.parent_gui.update_status()

            # Tell the workflow thread that the APs have been created successfully.
            self.signal_finished.emit()
        # Close the Window.
        self.close()


if __name__ == '__main__':
    # Images can either be extracted from a video file or a batch of single photographs. Select
    # the example for the test run.
    type = 'video'
    if type == 'image':
        names = glob('Images/2012*.tif')
        # names = glob.glob('Images/Moon_Tile-031*ap85_8b.tif')
        # names = glob.glob('Images/Example-3*.jpg')
    else:
        names = 'Videos/another_short_video.avi'
    print(names)

    # Get configuration parameters.
    configuration = Configuration()
    configuration.initialize_configuration()
    try:
        frames = Frames(configuration, names, type=type)
        print("Number of images read: " + str(frames.number))
        print("Image shape: " + str(frames.shape))
    except Error as e:
        print("Error: " + e.message)
        exit()

    # Rank the frames by their overall local contrast.
    rank_frames = RankFrames(frames, configuration)
    start = time()
    rank_frames.frame_score()
    end = time()
    print('Elapsed time in ranking images: {}'.format(end - start))
    print("Index of maximum: " + str(rank_frames.frame_ranks_max_index))
    print("Frame scores: " + str(rank_frames.frame_ranks))
    print("Frame scores (sorted): " + str(
        [rank_frames.frame_ranks[i] for i in rank_frames.quality_sorted_indices]))
    print("Sorted index list: " + str(rank_frames.quality_sorted_indices))

    # Initialize the frame alignment object.
    align_frames = AlignFrames(frames, rank_frames, configuration)

    if configuration.align_frames_mode == "Surface":
        start = time()
        # Select the local rectangular patch in the image where the L gradient is highest in both x
        # and y direction. The scale factor specifies how much smaller the patch is compared to the
        # whole image frame.
        (y_low_opt, y_high_opt, x_low_opt, x_high_opt) = align_frames.compute_alignment_rect(
            configuration.align_frames_rectangle_scale_factor)
        end = time()
        print('Elapsed time in computing optimal alignment rectangle: {}'.format(end - start))
        print("optimal alignment rectangle, x_low: " + str(x_low_opt) + ", x_high: " + str(
            x_high_opt) + ", y_low: " + str(y_low_opt) + ", y_high: " + str(y_high_opt))
        reference_frame_with_alignment_points = frames.frames_mono(
            rank_frames.frame_ranks_max_index).copy()
        reference_frame_with_alignment_points[y_low_opt,
        x_low_opt:x_high_opt] = reference_frame_with_alignment_points[y_high_opt - 1,
                                x_low_opt:x_high_opt] = 255
        reference_frame_with_alignment_points[y_low_opt:y_high_opt,
        x_low_opt] = reference_frame_with_alignment_points[y_low_opt:y_high_opt,
                     x_high_opt - 1] = 255
        # plt.imshow(reference_frame_with_alignment_points, cmap='Greys_r')
        # plt.show()

    # Align all frames globally relative to the frame with the highest score.
    start = time()
    try:
        align_frames.align_frames()
    except NotSupportedError as e:
        print("Error: " + e.message)
        exit()
    except InternalError as e:
        print("Warning: " + e.message)
        for index, frame_number in enumerate(align_frames.failed_index_list):
            print("Shift computation failed for frame " + str(
                align_frames.failed_index_list[index]) + ", minima list: " + str(
                align_frames.dev_r_list[index]))
    end = time()
    print('Elapsed time in aligning all frames: {}'.format(end - start))
    print("Frame shifts: " + str(align_frames.frame_shifts))
    print("Intersection: " + str(align_frames.intersection_shape))

    start = time()
    # Compute the reference frame by averaging the best frames.
    average = align_frames.average_frame()

    # Initialize the AlignmentPoints object.
    alignment_points = AlignmentPoints(configuration, frames, rank_frames, align_frames)
    end = time()
    print('Elapsed time in computing reference frame: {}'.format(end - start))
    print("Reference frame computed from the best " + str(
        align_frames.average_frame_number) + " frames.")
    # plt.imshow(align_frames.mean_frame, cmap='Greys_r')
    # plt.show()

    app = QtWidgets.QApplication(argv)
    window = AlignmentPointEditorWidget(None, configuration, align_frames, alignment_points, None)
    window.setMinimumSize(800,600)
    window.showMaximized()
    app.exec_()

    print("After AP editing, number of APs: " + str(
        len(alignment_points.alignment_points)) + ", aps dropped because too dim: " + str(
        alignment_points.alignment_points_dropped_dim) + ", aps dropped because too little "
                                                              "structure: " + str(
        alignment_points.alignment_points_dropped_structure))

    count_updates = 0
    for ap in alignment_points.alignment_points:
        if ap['reference_box'] is not None:
            continue
        count_updates += 1
        AlignmentPoints.set_reference_box(ap, align_frames.mean_frame)
    print("Buffers allocated for " + str(count_updates) + " alignment points.")
    exit()

